import logging
import math
import os
import pickle
import threading
import time
from collections import defaultdict
from collections import deque

import boto3

OBJECT_SIZE = 400000
MAX_SAVE_QUEUE = 500
log = logging.getLogger(__name__)


class DynamoDBStateStore(object):
    def __init__(self, name, dynamodb_region, stopping=False) -> None:
        self.dynamodb = boto3.resource('dynamodb', region_name=dynamodb_region)
        self.client = boto3.client('dynamodb', region_name=dynamodb_region)
        self.name = name
        self.dynamodb_region = dynamodb_region
        self.table = self.dynamodb.Table(name)
        self.stopping = stopping
        self.save_queue = deque()
        self.save_errors = 0
        self.save_thread = threading.Thread(target=self._save_loop, args=(), daemon=True)
        self.save_thread.start()

    def build_key(self, type, iden) -> str:
        """
        It builds a unique partition key. The key could be objects with __str__ method.
        """
        return f"{type} {iden}"

    def restore(self, keys) -> dict:
        """
        Fetch all under the same parition key(keys).
        ret: <dict of key to states>
        """
        first_items = self._get_first_partitions(keys)
        remaining_items = self._get_remaining_partitions(first_items)
        vals = self._merge_items(first_items, remaining_items)
        return vals

    def _get_items(self, keys: list) -> object:
        items = []
        for i in range(0, len(keys), 100):
            count = 0
            cand_keys = keys[i:min(len(keys), i + 100)]
            while True:
                resp = self.client.batch_get_item(
                    RequestItems={
                        self.name: {
                            'Keys': cand_keys,
                            'ConsistentRead': True
                        },
                    }
                )
                items.extend(resp['Responses'][self.name])
                if resp['UnprocessedKeys'].get(self.name) and count < 10:
                    cand_keys = resp['UnprocessedKeys'][self.name]['Keys']
                    count += 1
                elif count >= 10:
                    error = Exception(f'tron_dynamodb_restore_failure: failed to retrieve items with keys \n{cand_keys}\n from dynamodb\n{resp}')
                    raise error
                else:
                    break
        return items

    def _get_first_partitions(self, keys: list):
        new_keys = [{'key': {'S': key}, 'index': {'N': '0'}} for key in keys]
        return self._get_items(new_keys)

    def _get_remaining_partitions(self, items: list):
        keys_for_remaining_items = []
        for item in items:
            remaining_items = [{'key': {'S': str(item['key']['S'])}, 'index': {'N': str(i)}}
                               for i in range(1, int(item['num_partitions']['N']))]
            keys_for_remaining_items.extend(remaining_items)
        return self._get_items(keys_for_remaining_items)

    def _merge_items(self, first_items, remaining_items) -> dict:
        items = defaultdict(list)
        raw_items = defaultdict(bytearray)
        # Merge all items based their keys and deserialize their values
        if remaining_items:
            first_items.extend(remaining_items)
        for item in first_items:
            key = item['key']['S']
            items[key].append(item)
        for key, item in items.items():
            item.sort(key=lambda x: int(x['index']['N']))
            for val in item:
                raw_items[key] += bytes(val['val']['B'])
        deserialized_items = {k: pickle.loads(v) for k, v in raw_items.items()}
        return deserialized_items

    def save(self, key_value_pairs) -> None:
        for key, val in key_value_pairs:
            while True:
                qlen = len(self.save_queue)
                if qlen > MAX_SAVE_QUEUE:
                    log.info(f"save queue size {qlen} > {MAX_SAVE_QUEUE}, sleeping 5s")
                    time.sleep(5)
                    continue
                self.save_queue.append((key, pickle.dumps(val)))
                break

    def _consume_save_queue(self):
        qlen = len(self.save_queue)
        saved = 0
        start = time.time()
        for _ in range(qlen):
            try:
                key, val = self.save_queue.popleft()
                # Remove all previous data with the same partition key
                # TODO: only remove excess partitions if new data has fewer
                self._delete_item(key)
                self[key] = val
                # reset errors count if we can successfully save
                saved += 1
            except Exception as e:
                error = (
                    'tron_dynamodb_save_failure: failed to save key'
                    f'"{key}" to dynamodb:\n{repr(e)}'
                )
                log.error(error)
                self.save_queue.append((key, val))
        duration = time.time() - start
        log.info(f"saved {saved} items in {duration}s")

        if saved < qlen:
            self.save_errors += 1
        else:
            self.save_errors = 0

    def _save_loop(self):
        while True:
            if self.stopping:
                self._consume_save_queue()
                return

            if len(self.save_queue) == 0:
                log.debug("save queue empty, sleeping 5s")
                time.sleep(5)
                continue

            self._consume_save_queue()
            if self.save_errors > 100:
                log.error("too many dynamodb errors in a row, crashing")
                os.exit(1)

    def __setitem__(self, key: str, val: bytes) -> None:
        """
        Partition the item and write up to 10 partitions atomically.
        Retry up to 3 times on failure

        Examine the size of `val`, and splice it into
        different parts under 400KB with different sort keys,
        and save them under the same partition key built.
        """
        num_partitions = math.ceil(len(val) / OBJECT_SIZE)
        items = []
        for index in range(num_partitions):
            item = {
                'Put': {
                    'Item': {
                        'key': {
                            'S': key,
                        },
                        'index': {
                            'N': str(index),
                        },
                        'val': {
                            'B': val[index * OBJECT_SIZE:min(index * OBJECT_SIZE + OBJECT_SIZE, len(val))],
                        },
                        'num_partitions': {
                            'N': str(num_partitions),
                        }
                    },
                    'TableName': self.name,
                },
            }
            count = 0
            items.append(item)
            # Only up to 10 items are allowed per transactions
            while len(items) == 10 or index == num_partitions - 1:
                try:
                    self.client.transact_write_items(TransactItems=items)
                    items = []
                    break  # exit the while loop on successful writing
                except Exception as e:
                    count += 1
                    if count > 3:
                        raise e

    def _delete_item(self, key: str) -> None:
        with self.table.batch_writer() as batch:
            for index in range(self._get_num_of_partitions(key)):
                batch.delete_item(
                    Key={
                        'key': key,
                        'index': index,
                    }
                )

    def _get_num_of_partitions(self, key: str) -> int:
        """
        Return how many parts is the item partitioned into
        """
        try:
            partition = self.table.get_item(
                Key={
                    'key': key,
                    'index': 0,
                },
                ProjectionExpression='num_partitions',
                ConsistentRead=True
            )
            return int(partition.get('Item', {}).get('num_partitions', 0))
        except self.client.exceptions.ResourceNotFoundException:
            return 0

    def cleanup(self) -> None:
        self.stopping = True
        self.save_thread.join()
        return



#!/usr/bin/env python3.6
#
# get_tron_metrics.py
#   This script is designed to retrieve metrics from Tron via its API and send
#   send them to meteorite.
import logging
import pprint
import subprocess
import sys
import textwrap

from tron.commands import cmd_utils
from tron.commands.client import Client

log = logging.getLogger('get_tron_metrics')


def parse_cli():
    parser = cmd_utils.build_option_parser()
    parser.description = (
        "Collects metrics from Tron via its API and forwards them to "
        "meteorite."
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        default=False,
        help="Don't actually send metrics out. Defaults: %(default)s"
    )
    args = parser.parse_args()
    return args


def check_bin_exists(bin):
    """
    Checks if an executable binary exists

    :param bin: (str) Name of the executable; could be a path to one
    """
    return subprocess.call(
        ['which', bin],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) == 0


def send_data_metric(name, metric_type, value, dimensions={}, dry_run=False):
    """
    Sends a single data point to meteorite via bash command

    :param name: (str) Name of the metric
    :param metric_type: (str) Type of the meteorite metric. Must be in
        METEORITE_TYPES
    :param value: (float) Value of the metric
    :param dimensions: (dict) Metric dimensions as key-value pairs
    :param dry_run: (bool) Whether or not to send metrics to meteorite
    """
    if dry_run:
        metric_args = dict(
            name=name,
            metric_type=metric_type,
            value=value,
            dimensions=dimensions,
        )
        log.info(
            f"Would have sent this to meteorite:\n"
            f"{pprint.pformat(metric_args)}"
        )
        return

    cmd = ['meteorite', 'data', '-v', name, metric_type, str(value)]
    for k, v in dimensions.items():
        cmd.extend(['-d', f"{k}:{v}"])

    process = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    output, error = process.communicate()
    output = output.decode('utf-8').rstrip()
    error = error.decode('utf-8').rstrip()

    if process.returncode != 0:
        log.error(
            "Meteorite failed with:\n"
            f"{textwrap.indent(error, '    ')}"
        )
    else:
        log.debug(f"From meteorite: {output}")


def send_counter(name, **kwargs):
    send_data_metric(
        name=name,
        metric_type='counter',
        value=kwargs.pop('count'),
        dimensions=kwargs.pop('dimensions', {}),
        dry_run=kwargs.pop('dry_run', False),
    )


def send_gauge(name, **kwargs):
    send_data_metric(
        name=name,
        metric_type='gauge',
        value=kwargs.pop('value'),
        dimensions=kwargs.pop('dimensions', {}),
        dry_run=kwargs.pop('dry_run', False),
    )


def send_meter(name, **kwargs):
    send_counter(name, **kwargs)  # We ignore mX_rate args


def send_histogram(name, **kwargs):
    for k in ['p50', 'p75', 'p95', 'p99']:  # Only send p50-99
        gauge_name = f"{name}.{k}"
        kwargs['value'] = kwargs[k]  # set for gauge
        send_gauge(gauge_name, **kwargs)


def send_timer(name, **kwargs):
    # We mirror the metrics implementation in Tron by splitting timer into a
    # meter and a histogram
    send_meter(name, **kwargs)
    send_histogram(name, **kwargs)


_METRIC_SENDERS = {
    'counter': send_counter,
    'gauge': send_gauge,
    'meter': send_meter,
    'histogram': send_histogram,
    'timer': send_timer,
}


def send_metrics(metrics, cluster=None, dry_run=False):
    """
    Send metrics via meteorite

    :param metrics: Dictionary of metrics types and their data
    """
    for metric_type, data in metrics.items():
        for kwargs in data:
            name = kwargs.pop('name')
            kwargs['dry_run'] = dry_run

            if cluster:
                dimensions = kwargs.get('dimensions', {})
                dimensions['tron_cluster'] = cluster
                kwargs['dimensions'] = dimensions

            _METRIC_SENDERS[metric_type](name, **kwargs)


def main():
    args = parse_cli()
    cmd_utils.setup_logging(args)
    cmd_utils.load_config(args)
    client = Client(args.server, args.cluster_name)

    if check_bin_exists('meteorite'):
        metrics = client.metrics()
        send_metrics(metrics, cluster=client.cluster_name, dry_run=args.dry_run)
    else:
        log.error("'meteorite' was not found")


if __name__ == '__main__':
    sys.exit(main())



import json
import logging
import socket
import time
from urllib.parse import urlparse

import requests
import staticconf
from task_processing.runners.subscription import Subscription
from task_processing.task_processor import TaskProcessor
from twisted.internet.defer import logError

import tron.metrics as metrics
from tron.actioncommand import ActionCommand
from tron.utils.queue import PyDeferredQueue

TASK_LOG_FORMAT = '%(asctime)s %(name)s %(levelname)s %(message)s'
TASK_OUTPUT_LOGGER = 'tron.mesos.task_output'
CLUSTERMAN_YAML_FILE_PATH = "/nail/srv/configs/clusterman.yaml"
CLUSTERMAN_METRICS_YAML_FILE_PATH = "/nail/srv/configs/clusterman_metrics.yaml"

log = logging.getLogger(__name__)


def get_clusterman_metrics():
    try:
        import clusterman_metrics
        import clusterman_metrics.util.costs

        staticconf.YamlConfiguration(
            CLUSTERMAN_YAML_FILE_PATH, namespace="clusterman",
        )
        staticconf.YamlConfiguration(
            CLUSTERMAN_METRICS_YAML_FILE_PATH, namespace="clusterman_metrics",
        )
    except (ImportError, FileNotFoundError):
        clusterman_metrics = None

    return clusterman_metrics


def get_mesos_leader(master_address, mesos_master_port):
    url = "%s:%s/redirect" % (master_address, mesos_master_port)
    response = requests.get(url)
    return '{}:{}'.format(urlparse(response.url).hostname, mesos_master_port)


def combine_volumes(defaults, overrides):
    """Helper to reconcile lists of volume mounts.

    If any volumes have the same container path, the one in overrides wins.
    """
    result = {mount['container_path']: mount for mount in defaults}
    for mount in overrides:
        result[mount['container_path']] = mount
    return list(result.values())


def get_secret_from_file(file_path):
    if file_path is not None:
        with open(file_path) as f:
            secret = f.read().strip()
    else:
        secret = None
    return secret


class MesosClusterRepository:
    """A class that stores MesosCluster objects and configuration."""

    # Config values
    mesos_enabled = False
    master_address = None
    master_port = None
    secret_file = None
    role = None
    principal = None
    default_volumes = ()
    dockercfg_location = None
    offer_timeout = None
    secret = None

    name = 'frameworks'
    clusters = {}
    state_data = {}
    state_watcher = None

    @classmethod
    def attach(cls, _, observer):
        cls.state_watcher = observer

    @classmethod
    def get_cluster(cls, master_address=None):
        if master_address is None:
            master_address = cls.master_address
        if master_address not in cls.clusters:
            framework_id = cls.state_data.get(master_address)
            cluster = MesosCluster(
                mesos_address=master_address,
                mesos_master_port=cls.master_port,
                secret=cls.secret,
                principal=cls.principal,
                mesos_role=cls.role,
                framework_id=framework_id,
                enabled=cls.mesos_enabled,
                default_volumes=cls.default_volumes,
                dockercfg_location=cls.dockercfg_location,
                offer_timeout=cls.offer_timeout,
            )
            cls.clusters[master_address] = cluster
        return cls.clusters[master_address]

    @classmethod
    def shutdown(cls):
        for cluster in cls.clusters.values():
            cluster.stop()

    @classmethod
    def configure(cls, mesos_options):
        cls.master_address = mesos_options.master_address
        cls.master_port = mesos_options.master_port
        cls.secret_file = mesos_options.secret_file
        cls.role = mesos_options.role
        cls.secret = get_secret_from_file(cls.secret_file)
        cls.principal = mesos_options.principal
        cls.mesos_enabled = mesos_options.enabled
        cls.default_volumes = [
            vol._asdict() for vol in mesos_options.default_volumes
        ]
        cls.dockercfg_location = mesos_options.dockercfg_location
        cls.offer_timeout = mesos_options.offer_timeout

        for cluster in cls.clusters.values():
            cluster.set_enabled(cls.mesos_enabled)
            cluster.configure_tasks(
                default_volumes=cls.default_volumes,
                dockercfg_location=cls.dockercfg_location,
                offer_timeout=cls.offer_timeout,
            )

    @classmethod
    def restore_state(cls, mesos_state):
        cls.state_data = mesos_state.get(cls.name, {})

    @classmethod
    def save(cls, master_address, framework_id):
        cls.state_data[master_address] = framework_id
        cls.state_watcher.handler(cls, None)

    @classmethod
    def remove(cls, master_address):
        if master_address in cls.state_data:
            del cls.state_data[master_address]
            cls.state_watcher.handler(cls, None)


class MesosTask(ActionCommand):
    ERROR_STATES = frozenset(['failed', 'killed', 'error'])

    def __init__(self, id, task_config, serializer=None):
        super(MesosTask, self).__init__(id, task_config.cmd, serializer)
        self.task_config = task_config

        self.log = self.get_event_logger()
        self.setup_output_logging()

        self.log.info(
            'Mesos task {} created with config {}'.format(
                self.get_mesos_id(),
                self.get_config(),
            ),
        )

    def get_event_logger(self):
        log = logging.getLogger(__name__ + '.' + self.id)
        # Every time a task gets created, this function runs and will add
        # more stderr handlers to the logger, which results in duplicate log
        # output. We only want to add the stderr handler if the logger does not
        # have a handler yet.
        if not len(log.handlers):
            handler = logging.StreamHandler(self.stderr)
            handler.setFormatter(logging.Formatter(TASK_LOG_FORMAT))
            log.addHandler(handler)
        return log

    def setup_output_logging(self):
        task_id = self.get_mesos_id()
        stdout_logger = logging.getLogger(
            '{}.{}.{}'.format(TASK_OUTPUT_LOGGER, task_id, 'stdout'),
        )
        stdout_logger.addHandler(logging.StreamHandler(self.stdout))
        stderr_logger = logging.getLogger(
            '{}.{}.{}'.format(TASK_OUTPUT_LOGGER, task_id, 'stderr'),
        )
        stderr_logger.addHandler(logging.StreamHandler(self.stderr))

    def get_mesos_id(self):
        return self.task_config.task_id

    def get_config(self):
        return self.task_config

    def report_resources(self, decrement=False):
        multiplier = -1 if decrement else 1
        metrics.count(f'tron.mesos.cpus', self.task_config.cpus * multiplier)
        metrics.count(f'tron.mesos.mem', self.task_config.mem * multiplier)
        metrics.count(f'tron.mesos.disk', self.task_config.disk * multiplier)

    def log_event_info(self, event):
        # Separate out so task still transitions even if this nice-to-have logging fails.
        mesos_type = getattr(event, 'platform_type', None)
        if mesos_type == 'staging':
            # TODO: Save these in state?
            agent = event.raw.get('offer', {}).get('agent_id', {}).get('value')
            hostname = event.raw.get('offer', {}).get('hostname')
            self.log.info(
                f'Staging task on agent {agent} (hostname {hostname})'
            )
        elif mesos_type == 'running':
            agent = event.raw.get('agent_id', {}).get('value')
            self.log.info(f'Running on agent {agent}')
        elif mesos_type == 'finished':
            pass
        elif mesos_type in self.ERROR_STATES:
            self.log.error(f'Error from Mesos: {event.raw}')
        elif mesos_type is None:
            self.log.info(f'Non-Mesos event: {event.raw}')
            if 'Failed due to offer timeout' in str(event.raw):
                self.log.info(f'Explanation:')
                self.log.info(f'This error means that Tron timed out waiting for Mesos to give it the')
                self.log.info(f'resources requested (ram, cpu, disk, pool, etc).')
                self.log.info(f'This can happen if the cluster is low on resources, or if the resource')
                self.log.info(f'requests are too high.')
                self.log.info(f'Try reducing the resource request, or adding retries + retries_delay.')
                self.log.info(f'')

        # Mesos events may have task reasons
        if mesos_type:
            message = event.raw.get('message', '')
            reason = event.raw.get('reason', '')
            if message or reason:
                self.log.info(f'More info: {reason}: {message}')

    def handle_event(self, event):
        event_id = getattr(event, 'task_id', None)
        if event_id != self.get_mesos_id():
            self.log.warning(
                'Event task id {} does not match, ignoring'.format(event_id),
            )
            return
        mesos_type = getattr(event, 'platform_type', None)

        self.log.info(
            'Got event for task {id}, Mesos type {type}'.format(
                id=event_id,
                type=mesos_type,
            )
        )
        try:
            self.log_event_info(event)
        except Exception as e:
            self.log.warning('Exception while logging event: {}'.format(e))

        if mesos_type == 'staging':
            pass
        elif mesos_type == 'starting':
            self.started()
        elif mesos_type == 'running':
            self.started()
        elif mesos_type == 'finished':
            self.exited(0)
        elif mesos_type == 'lost':
            self.log.warning("Mesos does not know anything about this task, it is LOST")
            self.log.warning("This can happen for any number of reasons, and Tron can't know if the task ran or not at all!")
            self.log.warning("If you want Tron to RUN it (again) anyway, retry it with:")
            self.log.warning(f"    tronctl retry {self.id}")
            self.log.warning("If you want Tron to NOT run it and consider it as a success, skip it with:")
            self.log.warning(f"    tronctl skip {self.id}")
            self.log.warning("If you want Tron to NOT run it and consider it as a failure, fail it with:")
            self.log.warning(f"    tronctl fail {self.id}")
            self.exited(None)
        elif mesos_type in self.ERROR_STATES:
            self.exited(1)
        elif mesos_type is None:
            pass
        else:
            self.log.info(
                'Did not handle unknown mesos event type: {}'.format(event),
            )

        if event.terminal:
            self.log.info('This Mesos event was terminal, ending this action')
            self.report_resources(decrement=True)

            exit_code = int(not getattr(event, 'success', False))
            # Returns False if we've already exited normally above
            unexpected_error = self.exited(exit_code)
            if unexpected_error:
                self.log.error('Unexpected failure, exiting')

            self.done()


class MesosCluster:
    def __init__(
        self,
        mesos_address,
        mesos_master_port=None,
        secret=None,
        principal=None,
        mesos_role=None,
        framework_id=None,
        enabled=True,
        default_volumes=None,
        dockercfg_location=None,
        offer_timeout=None,
    ):
        self.mesos_address = mesos_address
        self.mesos_master_port = mesos_master_port
        self.secret = secret
        self.principal = principal
        self.mesos_role = mesos_role
        self.enabled = enabled
        self.default_volumes = default_volumes or []
        self.dockercfg_location = dockercfg_location
        self.offer_timeout = offer_timeout
        self.framework_id = framework_id

        self.processor = TaskProcessor()
        self.queue = PyDeferredQueue()
        self.deferred = None
        self.runner = None
        self.tasks = {}

        self.processor.load_plugin(
            provider_module='task_processing.plugins.mesos'
        )
        self.connect()

    def set_enabled(self, is_enabled):
        self.enabled = is_enabled
        if is_enabled:
            self.connect()
        else:
            self.stop(fail_tasks=True)

    def configure_tasks(
        self,
        default_volumes,
        dockercfg_location,
        offer_timeout,
    ):
        self.default_volumes = default_volumes
        self.dockercfg_location = dockercfg_location
        self.offer_timeout = offer_timeout

    def connect(self):
        self.runner = self.get_runner(self.mesos_address, self.queue)
        self.handle_next_event()

    def handle_next_event(self, deferred_result=None):
        if self.deferred and not self.deferred.called:
            log.warning(
                'Already have handlers waiting for next event in queue, '
                'not adding more'
            )
            return
        self.deferred = self.queue.get()
        self.deferred.addCallback(self._process_event)
        self.deferred.addCallback(self.handle_next_event)
        self.deferred.addErrback(logError)
        self.deferred.addErrback(self.handle_next_event)

    def _check_connection(self):
        if self.runner.stopping:
            # Last framework was terminated for some reason, re-connect.
            log.info('Last framework stopped, re-connecting')
            self.connect()
        elif self.deferred.called:
            # Just in case callbacks are missing, re-add.
            self.handle_next_event()

    def submit(self, task):
        if not task:
            return

        if not self.enabled:
            task.log.info('Task failed to start, Mesos is disabled.')
            task.exited(1)
            return
        self._check_connection()

        mesos_task_id = task.get_mesos_id()
        self.tasks[mesos_task_id] = task
        env = task.get_config()['environment']
        clusterman_resource_str = env.get('CLUSTERMAN_RESOURCES')
        clusterman_metrics = get_clusterman_metrics()
        if clusterman_resource_str and clusterman_metrics:
            clusterman_resources = json.loads(clusterman_resource_str)
            cluster = env.get('EXECUTOR_CLUSTER', env.get('PAASTA_CLUSTER'))
            pool = env.get('EXECUTOR_POOL', env.get('PAASTA_POOL'))
            aws_region = staticconf.read(f'clusters.{cluster}.aws_region', namespace='clusterman')
            metrics_client = clusterman_metrics.ClustermanMetricsBotoClient(
                region_name=aws_region,
                app_identifier=pool,
            )
            with metrics_client.get_writer(
                clusterman_metrics.APP_METRICS, aggregate_meteorite_dims=True
            ) as writer:
                for metric_key, metric_value in clusterman_resources.items():
                    writer.send((metric_key, int(time.time()), metric_value))
        self.runner.run(task.get_config())
        log.info(
            'Submitting task {} to {}'.format(
                mesos_task_id,
                self.mesos_address,
            ),
        )
        task.report_resources()

    def recover(self, task):
        if not task:
            return

        if not self.enabled:
            task.log.info('Could not recover task, Mesos is disabled.')
            task.exited(None)
            return
        self._check_connection()

        mesos_task_id = task.get_mesos_id()
        self.tasks[mesos_task_id] = task
        task.log.info('TRON RESTARTED! Starting recovery procedure by reconciling state for this task from Mesos')
        task.started()
        self.runner.reconcile(task.get_config())
        task.report_resources()

    def create_task(
        self,
        action_run_id,
        command,
        cpus,
        mem,
        disk,
        constraints,
        docker_image,
        docker_parameters,
        env,
        extra_volumes,
        serializer,
        task_id=None,
    ):
        if not self.runner:
            return None

        uris = [self.dockercfg_location] if self.dockercfg_location else []
        volumes = combine_volumes(self.default_volumes, extra_volumes)
        task_kwargs = {
            'name': action_run_id,
            'cmd': command,
            'cpus': cpus,
            'mem': mem,
            'disk': disk,
            'constraints': constraints,
            'image': docker_image,
            'docker_parameters': docker_parameters,
            'environment': env,
            'volumes': volumes,
            'uris': uris,
            'offer_timeout': self.offer_timeout,
        }
        task_config = self.runner.TASK_CONFIG_INTERFACE(**task_kwargs)

        if task_id is not None:
            try:
                task_config = task_config.set_task_id(task_id)
            except ValueError:
                log.error(f'Invalid {task_id} for {action_run_id}')
                return

        return MesosTask(action_run_id, task_config, serializer)

    def get_runner(self, mesos_address, queue):
        if not self.enabled:
            log.info('Mesos is disabled, not creating a framework.')
            return None

        if self.runner and not self.runner.stopping:
            log.info('Already have a running framework, not creating one.')
            return self.runner

        framework_name = 'tron-{}'.format(socket.gethostname())
        executor = self.processor.executor_from_config(
            provider='mesos_task',
            provider_config={
                'secret':
                    self.secret,
                'principal':
                    self.principal,
                'mesos_address':
                    get_mesos_leader(mesos_address, self.mesos_master_port),
                'role':
                    self.mesos_role,
                'framework_name':
                    framework_name,
                'framework_id':
                    self.framework_id,
                'failover':
                    True,
            }
        )

        def log_output(task_id, message, stream):
            logger = logging.getLogger(
                '{}.{}.{}'.format(
                    TASK_OUTPUT_LOGGER,
                    task_id,
                    stream,
                )
            )
            logger.info(message)

        logging_executor = self.processor.executor_from_config(
            provider='logging',
            provider_config={
                'downstream_executor': executor,
                'handler': log_output,
                'format_string': '{line}',
            },
        )
        return Subscription(logging_executor, queue)

    def _process_event(self, event):
        if event.kind == 'control':
            message = getattr(event, 'message', None)
            if message == 'stop':
                # Framework has been removed, stop it.
                log.warning('Framework has been stopped: {}'.format(event.raw))
                self.stop()
                MesosClusterRepository.remove(self.mesos_address)
            elif message == 'unknown':
                log.warning(
                    'Unknown error from Mesos master: {}'.format(event.raw)
                )
            elif message == 'registered':
                framework_id = event.raw['framework_id']['value']
                MesosClusterRepository.save(self.mesos_address, framework_id)
            else:
                log.warning('Unknown type of control event: {}'.format(event))

        elif event.kind == 'task':
            if not hasattr(event, 'task_id'):
                log.warning('Task event missing task_id: {}'.format(event))
                return
            if event.task_id not in self.tasks:
                log.warning(
                    'Received event for unknown task {}: {}'.format(
                        event.task_id,
                        event,
                    ),
                )
                return
            task = self.tasks[event.task_id]
            task.handle_event(event)
            if task.is_done:
                del self.tasks[event.task_id]
        else:
            log.warning('Unknown type of event: {}'.format(event))

    def stop(self, fail_tasks=False):
        self.framework_id = None
        if self.runner:
            self.runner.stop()

        # Clear message queue
        if self.deferred:
            self.deferred.cancel()
            self.deferred = None
        self.queue = PyDeferredQueue()

        if fail_tasks:
            for key, task in list(self.tasks.items()):
                task.exited(None)
                del self.tasks[key]

    def kill(self, task_id):
        return self.runner.kill(task_id)



from unittest import TestCase

from testifycompat.assertions import *
from testifycompat.fixtures import *


version = '0.1.2'


def run():
    raise AssertionError(
        "Oops, you tried to use testifycompat.run(). This function doesn't "
        "do anything, it only exists as backwards compatibility with testify. "
        "You should remove it from your code.")



from __future__ import absolute_import
from __future__ import unicode_literals

import datetime

import pytz

from testifycompat import assert_equal
from testifycompat import run
from testifycompat import TestCase
from tron.utils import trontimespec


class TestGetTime(TestCase):
    def test_get_time(self):
        assert_equal(datetime.time(4, 15), trontimespec.get_time("4:15"))
        assert_equal(datetime.time(22, 59), trontimespec.get_time("22:59"))

    def test_get_time_invalid_time(self):
        assert not trontimespec.get_time("25:00")
        assert not trontimespec.get_time("22:61")


class TestTimeSpecification(TestCase):
    def _cmp(self, start_time, expected):
        start_time = datetime.datetime(*start_time)
        expected = datetime.datetime(*expected)
        assert_equal(self.time_spec.get_match(start_time), expected)

    def test_get_match_months(self):
        self.time_spec = trontimespec.TimeSpecification(months=[1, 5])
        self._cmp((2012, 3, 14), (2012, 5, 1))
        self._cmp((2012, 5, 22), (2012, 5, 23))
        self._cmp((2012, 12, 22), (2013, 1, 1))

    def test_get_match_monthdays(self):
        self.time_spec = trontimespec.TimeSpecification(
            monthdays=[10, 3, 3, 10],
        )
        self._cmp((2012, 3, 14), (2012, 4, 3))
        self._cmp((2012, 3, 1), (2012, 3, 3))

    def test_get_match_weekdays(self):
        self.time_spec = trontimespec.TimeSpecification(weekdays=[2, 3])
        self._cmp((2012, 3, 14), (2012, 3, 20))
        self._cmp((2012, 3, 20), (2012, 3, 21))

    def test_next_month_generator(self):
        time_spec = trontimespec.TimeSpecification(months=[2, 5])
        gen = time_spec.next_month(datetime.datetime(2012, 3, 14))
        expected = [(5, 2012), (2, 2013), (5, 2013), (2, 2014)]
        assert_equal([next(gen) for _ in range(4)], expected)

    def test_next_day_monthdays(self):
        time_spec = trontimespec.TimeSpecification(monthdays=[5, 10, 15])
        gen = time_spec.next_day(14, 2012, 3)
        assert_equal(list(gen), [15])

        gen = time_spec.next_day(1, 2012, 3)
        assert_equal(list(gen), [5, 10, 15])

    def test_next_day_monthdays_with_last(self):
        time_spec = trontimespec.TimeSpecification(monthdays=[5, 'LAST'])
        gen = time_spec.next_day(14, 2012, 3)
        assert_equal(list(gen), [31])

    def test_next_day_weekdays(self):
        time_spec = trontimespec.TimeSpecification(weekdays=[1, 5])
        gen = time_spec.next_day(14, 2012, 3)
        assert_equal(list(gen), [16, 19, 23, 26, 30])

        gen = time_spec.next_day(1, 2012, 3)
        assert_equal(list(gen), [2, 5, 9, 12, 16, 19, 23, 26, 30])

    def test_next_day_weekdays_with_ordinals(self):
        time_spec = trontimespec.TimeSpecification(
            weekdays=[1, 5],
            ordinals=[1, 3],
        )
        gen = time_spec.next_day(14, 2012, 3)
        assert_equal(list(gen), [16, 19])

        gen = time_spec.next_day(1, 2012, 3)
        assert_equal(list(gen), [2, 5, 16, 19])

    def test_next_time_timestr(self):
        time_spec = trontimespec.TimeSpecification(timestr="13:13")
        start_date = datetime.datetime(2012, 3, 14, 0, 15)
        time = time_spec.next_time(start_date, True)
        assert_equal(time, datetime.time(13, 13))

        start_date = datetime.datetime(2012, 3, 14, 13, 13)
        assert time_spec.next_time(start_date, True) is None
        time = time_spec.next_time(start_date, False)
        assert_equal(time, datetime.time(13, 13))

    def test_next_time_hours(self):
        time_spec = trontimespec.TimeSpecification(hours=[4, 10])
        start_date = datetime.datetime(2012, 3, 14, 0, 15)
        time = time_spec.next_time(start_date, True)
        assert_equal(time, datetime.time(4, 0))

        start_date = datetime.datetime(2012, 3, 14, 13, 13)
        assert time_spec.next_time(start_date, True) is None
        time = time_spec.next_time(start_date, False)
        assert_equal(time, datetime.time(4, 0))

    def test_next_time_minutes(self):
        time_spec = trontimespec.TimeSpecification(
            minutes=[30, 20, 30],
            seconds=[0],
        )
        start_date = datetime.datetime(2012, 3, 14, 0, 25)
        time = time_spec.next_time(start_date, True)
        assert_equal(time, datetime.time(0, 30))

        start_date = datetime.datetime(2012, 3, 14, 23, 30)
        assert time_spec.next_time(start_date, True) is None
        time = time_spec.next_time(start_date, False)
        assert_equal(time, datetime.time(0, 20))

    def test_next_time_hours_and_minutes_and_seconds(self):
        time_spec = trontimespec.TimeSpecification(
            minutes=[20, 30],
            hours=[1, 5],
            seconds=[4, 5],
        )
        start_date = datetime.datetime(2012, 3, 14, 1, 25)
        time = time_spec.next_time(start_date, True)
        assert_equal(time, datetime.time(1, 30, 4))

        start_date = datetime.datetime(2012, 3, 14, 5, 30, 6)
        assert time_spec.next_time(start_date, True) is None
        time = time_spec.next_time(start_date, False)
        assert_equal(time, datetime.time(1, 20, 4))

    def test_get_match_dst_spring_forward(self):
        tz = pytz.timezone('US/Pacific')
        time_spec = trontimespec.TimeSpecification(
            hours=[0, 1, 2, 3, 4],
            minutes=[0],
            seconds=[0],
            timezone='US/Pacific',
        )
        start = trontimespec.naive_as_timezone(datetime.datetime(2020, 3, 8, 1), tz)
        # Springing forward, the next hour after 1AM should be 3AM
        next_time = time_spec.get_match(start)
        assert next_time.hour == 3

    def test_get_match_dst_fall_back(self):
        tz = pytz.timezone('US/Pacific')
        time_spec = trontimespec.TimeSpecification(
            hours=[0, 1, 2, 3, 4],
            minutes=[0],
            seconds=[0],
            timezone='US/Pacific',
        )
        start = trontimespec.naive_as_timezone(datetime.datetime(2020, 11, 1, 1), tz)
        # Falling back, the next hour after 1AM is 1AM again. But we only run on the first 1AM
        # Next run time should be 2AM
        next_time = time_spec.get_match(start)
        assert next_time.hour == 2


if __name__ == "__main__":
    run()



from __future__ import absolute_import
from __future__ import unicode_literals

from unittest import mock

from testifycompat import assert_equal
from testifycompat import run
from testifycompat import setup
from testifycompat import TestCase
from tests.assertions import assert_length
from tron.utils.observer import Observable
from tron.utils.observer import Observer


class TestObservable(TestCase):
    @setup
    def setup_observer(self):
        self.obs = Observable()

    def test_attach(self):
        def func():
            return 1

        self.obs.attach('a', func)
        assert_equal(len(self.obs._observers), 1)
        assert_equal(self.obs._observers['a'], [func])

    def test_listen_seq(self):
        def func():
            return 1

        self.obs.attach(['a', 'b'], func)
        assert_equal(len(self.obs._observers), 2)
        assert_equal(self.obs._observers['a'], [func])
        assert_equal(self.obs._observers['b'], [func])

    def test_notify(self):
        handler = mock.MagicMock()
        self.obs.attach(['a', 'b'], handler)
        self.obs.notify('a')
        assert_equal(len(handler.handler.mock_calls), 1)
        self.obs.notify('b')
        assert_equal(len(handler.handler.mock_calls), 2)


class TestObserverClear(TestCase):
    @setup
    def setup_observer(self):
        self.obs = Observable()

        def func():
            return 1

        self.obs.attach('a', func)
        self.obs.attach('b', func)
        self.obs.attach(True, func)
        self.obs.attach(['a', 'b'], func)

    def test_clear_listeners_all(self):
        self.obs.clear_observers()
        assert_equal(len(self.obs._observers), 0)

    def test_clear_listeners_some(self):
        self.obs.clear_observers('a')
        assert_equal(len(self.obs._observers), 2)
        assert_equal(set(self.obs._observers.keys()), {True, 'b'})

    def test_remove_observer_none(self):
        def observer():
            return 2

        self.obs.remove_observer(observer)
        assert_equal(set(self.obs._observers.keys()), {True, 'a', 'b'})
        assert_length(self.obs._observers['a'], 2)
        assert_length(self.obs._observers['b'], 2)
        assert_length(self.obs._observers[True], 1)

    def test_remove_observer(self):
        def observer():
            return 2

        self.obs.attach('a', observer)
        self.obs.attach('c', observer)
        self.obs.remove_observer(observer)
        assert_length(self.obs._observers['a'], 2)
        assert_length(self.obs._observers['b'], 2)
        assert_length(self.obs._observers[True], 1)
        assert_length(self.obs._observers['c'], 0)


class MockObserver(Observer):
    def __init__(self, obs, event):
        self.obs = obs
        self.event = event
        self.watch(obs, event)
        self.has_watched = 0

    def handler(self, obs, event):
        assert_equal(obs, self.obs)
        assert_equal(event, self.event)
        self.has_watched += 1


class TestObserver(TestCase):
    @setup
    def setup_observer(self):
        self.obs = Observable()

    def test_watch(self):
        event = "FIVE"
        handler = MockObserver(self.obs, event)

        self.obs.notify(event)
        assert_equal(handler.has_watched, 1)
        self.obs.notify("other event")
        assert_equal(handler.has_watched, 1)
        self.obs.notify(event)
        assert_equal(handler.has_watched, 2)


if __name__ == "__main__":
    run()



import hashlib
import logging
import os

from tron import yaml
from tron.config import config_parse
from tron.config import ConfigError
from tron.config import schema
from tron.core.jobgraph import JobGraph
from tron.utils import maybe_decode
from tron.utils import maybe_encode

log = logging.getLogger(__name__)


def from_string(content):
    try:
        return yaml.safe_load(content)
    except yaml.yaml.error.YAMLError as e:
        raise ConfigError("Invalid config format: %s" % str(e))


def write(path, content):
    with open(path, 'w') as fh:
        yaml.dump(content, fh)


def read(path):
    with open(path, 'r') as fh:
        return from_string(fh)


def write_raw(path, content):
    with open(path, 'w') as fh:
        fh.write(maybe_decode(content))


def read_raw(path):
    with open(path, 'r') as fh:
        return fh.read()


def hash_digest(content):
    return hashlib.sha1(maybe_encode(content)).hexdigest()


class ManifestFile(object):
    """Manage the manifest file, which tracks name to filename."""

    MANIFEST_FILENAME = '_manifest.yaml'

    def __init__(self, path):
        self.filename = os.path.join(path, self.MANIFEST_FILENAME)

    def create(self):
        if os.path.isfile(self.filename):
            msg = "Refusing to create manifest. File %s exists."
            log.info(msg % self.filename)
            return

        write(self.filename, {})

    def add(self, name, filename):
        manifest = read(self.filename)
        manifest[name] = filename
        write(self.filename, manifest)

    def delete(self, name):
        manifest = read(self.filename)
        if name not in manifest:
            msg = "Namespace %s does not exist in manifest, cannot delete."
            log.info(msg % name)
            return

        del manifest[name]
        write(self.filename, manifest)

    def get_file_mapping(self):
        return read(self.filename)

    def get_file_name(self, name):
        return self.get_file_mapping().get(name)

    def __contains__(self, name):
        return name in self.get_file_mapping()


class ConfigManager(object):
    """Read, load and write configuration."""

    DEFAULT_HASH = hash_digest("")

    def __init__(self, config_path, manifest=None):
        self.config_path = config_path
        self.manifest = manifest or ManifestFile(config_path)

    def build_file_path(self, name):
        name = name.replace('.', '_').replace(os.path.sep, '_')
        return os.path.join(self.config_path, '%s.yaml' % name)

    def read_raw_config(self, name=schema.MASTER_NAMESPACE):
        """Read the config file without converting to yaml."""
        filename = self.manifest.get_file_name(name)
        return read_raw(filename)

    def write_config(self, name, content):
        self.validate_with_fragment(
            name,
            from_string(content),
            # TODO: remove this constraint after tron triggers across clusters are supported.
            should_validate_missing_dependency=False,
        )
        filename = self.get_filename_from_manifest(name)
        write_raw(filename, content)

    def delete_config(self, name):
        filename = self.manifest.get_file_name(name)
        if not filename:
            msg = "Namespace %s does not exist in manifest, cannot delete."
            log.info(msg % name)
            return

        self.manifest.delete(name)
        os.remove(filename)

    def get_filename_from_manifest(self, name):
        def create_filename():
            filename = self.build_file_path(name)
            self.manifest.add(name, filename)
            return filename

        return self.manifest.get_file_name(name) or create_filename()

    def validate_with_fragment(
        self,
        name,
        content,
        should_validate_missing_dependency=True,
    ):
        name_mapping = self.get_config_name_mapping()
        name_mapping[name] = content
        try:
            JobGraph(
                config_parse.ConfigContainer.create(name_mapping),
                should_validate_missing_dependency=should_validate_missing_dependency,
            )
        except ValueError as e:
            raise ConfigError(str(e))

    def get_config_name_mapping(self):
        seq = self.manifest.get_file_mapping().items()
        return {name: read(filename) for name, filename in seq}

    def load(self):
        """Return the fully constructed configuration."""
        log.info("Loading full config from %s" % self.config_path)
        name_mapping = self.get_config_name_mapping()
        return config_parse.ConfigContainer.create(name_mapping)

    def get_hash(self, name):
        """Return a hash of the configuration contents for name."""
        if name not in self:
            return self.DEFAULT_HASH
        return hash_digest(self.read_raw_config(name))

    def __contains__(self, name):
        return name in self.manifest

    def get_namespaces(self):
        return self.manifest.get_file_mapping().keys()


def create_new_config(path, master_content):
    """Create a new configuration directory with master config."""
    os.makedirs(path)
    manager = ConfigManager(path)
    manager.manifest.create()
    filename = manager.get_filename_from_manifest(schema.MASTER_NAMESPACE)
    write_raw(filename, master_content)



import logging

from tron.core.job import Job
from tron.utils import collections
from tron.utils import proxy

log = logging.getLogger(__name__)


class JobCollection:
    """A collection of jobs."""

    def __init__(self):
        self.jobs = collections.MappingCollection('jobs')
        self.proxy = proxy.CollectionProxy(
            self.jobs.values,
            [
                proxy.func_proxy('enable', lambda seq: all(list(seq))),
                proxy.func_proxy('disable', lambda seq: all(list(seq))),
                proxy.func_proxy('schedule', lambda seq: all(list(seq))),
                proxy.func_proxy('run_queue_schedule', lambda seq: all(list(seq))),
            ],
        )

    def load_from_config(self, job_configs, factory, reconfigure):
        """Apply a configuration to this collection and return a generator of
        jobs which were added.
        """
        self.jobs.filter_by_name(job_configs)

        def map_to_job_and_schedule(job_schedulers):
            for job_scheduler in job_schedulers:
                if reconfigure:
                    job_scheduler.schedule()
                yield job_scheduler.get_job()

        seq = (factory.build(config) for config in job_configs.values())
        return map_to_job_and_schedule(filter(self.add, seq))

    def add(self, job_scheduler):
        return self.jobs.add(job_scheduler, self.update)

    def move(self, old_name, new_name):
        job_scheduler = self.get_by_name(old_name)

        # check if job is running
        if job_scheduler.get_job().status == Job.STATUS_RUNNING:
            return f"Moving {old_name} to {new_name} failed. Job is still running."

        log.info(f"Moving {old_name} to {new_name}")
        job_scheduler.update_name(new_name)
        self.add(self.jobs.pop(old_name))

        return f"Moving {old_name} to {new_name} succeeded."

    def update(self, new_job_scheduler):
        log.info(f"Updating {new_job_scheduler}")
        job_scheduler = self.get_by_name(new_job_scheduler.get_name())
        job_scheduler.update_from_job_scheduler(new_job_scheduler)
        job_scheduler.schedule_reconfigured()
        return True

    def restore_state(self, job_state_data, config_action_runner):
        for name, state in job_state_data.items():
            self.jobs[name].restore_state(state, config_action_runner)
        log.info(f"Loaded state for {len(job_state_data)} jobs")

    def get_by_name(self, name):
        return self.jobs.get(name)

    def get_names(self):
        return self.jobs.keys()

    def get_jobs(self):
        return [sched.get_job() for sched in self]

    def get_job_run_collections(self):
        return [sched.get_job_runs() for sched in self]

    def __iter__(self):
        for val in self.jobs.values():
            yield val

    def __getattr__(self, name):
        return self.proxy.perform(name)

    def __contains__(self, name):
        return name in self.jobs



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A complete time specification based on the Google App Engine GROC spec."""
import calendar
import datetime

import pytz


def get_timezone(timezone_string):
    """Converts a timezone string to a pytz timezone object.

    Arguments:
      timezone_string: a string representing a timezone, or None

    Returns:
      a pytz timezone object, or None if the input timezone_string is None

    Raises:
      ValueError: if timezone_string is not None and the pytz module could not be
          loaded
    """
    if timezone_string:
        return pytz.timezone(timezone_string)
    else:
        return None


def to_timezone(t, tzinfo):
    """Converts 't' to the time zone 'tzinfo'.

    Arguments:
      t: a datetime object.  It may be in any pytz time zone, or it may be
          timezone-naive (interpreted as UTC).
      tzinfo: a pytz timezone object, or None (interpreted as UTC).

    Returns:
      a datetime object in the time zone 'tzinfo'
    """
    if tzinfo:
        if not t.tzinfo:
            t = pytz.utc.localize(t)
        return tzinfo.normalize(t.astimezone(tzinfo))
    elif t.tzinfo:
        return pytz.utc.normalize(t.astimezone(pytz.utc)).replace(tzinfo=None)
    else:
        return t


def naive_as_timezone(t, tzinfo):
    """Interprets the naive datetime with the given time zone."""
    try:
        result = tzinfo.localize(t, is_dst=None)
    except pytz.AmbiguousTimeError:
        # We are in the infamous 1 AM block which happens twice on
        # fall-back. Pretend like it's the first time, every time.
        result = tzinfo.localize(t, is_dst=True)
    except pytz.NonExistentTimeError:
        # We are in the infamous 2:xx AM block which does not
        # exist. Pretend like it's the later time, every time.
        result = tzinfo.localize(t, is_dst=False)
    return result


def get_time(time_string):
    """Converts a string to a datetime.time object.

    Arguments:
      time_string: a string representing a time ('hours:minutes')

    Returns:
      a datetime.time object
    """
    try:
        return datetime.datetime.strptime(time_string, "%H:%M").time()
    except ValueError:
        return None


TOKEN_LAST = 'LAST'

ordinal_range = range(1, 6)
weekday_range = range(0, 7)
month_range = range(1, 13)
monthday_range = range(1, 32)
hour_range = range(0, 24)
minute_range = second_range = range(0, 60)


def validate_spec(source, value_range, type, default=None, allow_last=False):
    default = default if default is not None else value_range
    if not source:
        return default

    has_last = False
    source_wo_last = []
    for item in source:
        if allow_last and item == TOKEN_LAST:
            has_last = True
            continue
        if item not in value_range:
            raise ValueError("%s not in range %s" % (type, value_range))
        source_wo_last.append(item)

    sorted_source = sorted(source_wo_last)
    if has_last:
        sorted_source.append(TOKEN_LAST)

    return sorted_source


class TimeSpecification(object):
    """TimeSpecification determines the next time which matches the
    configured pattern.
    """

    def __init__(
        self,
        ordinals=None,
        weekdays=None,
        months=None,
        monthdays=None,
        timestr=None,
        timezone=None,
        minutes=None,
        hours=None,
        seconds=None,
    ):

        if weekdays and monthdays:
            raise ValueError('cannot supply both monthdays and weekdays')

        if timestr and (minutes or hours or seconds):
            raise ValueError('cannot supply both timestr and h/m/s')

        if not any((timestr, minutes, hours, seconds)):
            timestr = '00:00'

        if timestr:
            time = get_time(timestr)
            hours = [time.hour]
            minutes = [time.minute]
            seconds = [0]

        self.hours = validate_spec(hours, hour_range, 'hour')
        self.minutes = validate_spec(minutes, minute_range, 'minute')
        self.seconds = validate_spec(seconds, second_range, 'second')
        self.ordinals = validate_spec(ordinals, ordinal_range, 'ordinal')
        self.weekdays = validate_spec(
            weekdays,
            weekday_range,
            'weekdays',
            allow_last=True,
        )
        self.months = validate_spec(months, month_range, 'month')
        self.monthdays = validate_spec(
            monthdays,
            monthday_range,
            'monthdays',
            [],
            True,
        )
        self.timezone = get_timezone(timezone)

    def next_day(self, first_day, year, month):
        """Returns matching days for the given year and month.
        """
        first_day_of_month, last_day_of_month = calendar.monthrange(
            year,
            month,
        )

        def map_last(day, ):
            return last_day_of_month if day == TOKEN_LAST else day

        def day_filter(day):
            return first_day <= day <= last_day_of_month

        def sort_days(days):
            return sorted(filter(day_filter, days))

        if self.monthdays:
            return sort_days(map_last(day) for day in self.monthdays)

        start_day = (first_day_of_month + 1) % 7

        def days_from_weekdays():
            for ordinal in self.ordinals:
                week = (ordinal - 1) * 7
                for weekday in self.weekdays:
                    yield ((weekday - start_day) % 7) + week + 1

        return sort_days(days_from_weekdays())

    def next_month(self, start_date):
        """Create a generator which yields valid months after the start month.
        """
        current = start_date.month
        potential = [m for m in self.months if m >= current]
        year_wraps = 0

        while True:
            if not potential:
                year_wraps += 1
                potential = list(self.months)

            yield potential.pop(0), start_date.year + year_wraps

    def next_time(self, start_date, is_start_day):
        """Return the next valid time."""
        start_hour = start_date.time().hour

        def hour_filter(hour):
            return not is_start_day or hour >= start_hour

        for hour in filter(hour_filter, self.hours):
            for minute in self.minutes:
                for second in self.seconds:
                    candidate = datetime.time(hour, minute, second)

                    if is_start_day and start_date.time() >= candidate:
                        continue

                    return candidate

    def get_match(self, start):
        """Returns the next datetime match after start."""
        start_date = to_timezone(start, self.timezone).replace(tzinfo=None)

        def get_first_day(month, year):
            if (month, year) != (start_date.month, start_date.year):
                return 1
            return start_date.day

        for month, year in self.next_month(start_date):
            first_day = get_first_day(month, year)

            for day in self.next_day(first_day, year, month):
                is_start_day = start_date.timetuple()[:3] == (year, month, day)

                time = self.next_time(start_date, is_start_day)
                if time is None:
                    continue

                candidate = start_date.replace(
                    year,
                    month,
                    day,
                    time.hour,
                    time.minute,
                    second=time.second,
                    microsecond=0,
                )
                candidate = self.handle_timezone(candidate, start.tzinfo)
                if not candidate:
                    continue
                return candidate

    # TODO: test
    def handle_timezone(self, out, tzinfo):
        if self.timezone:
            out = naive_as_timezone(out, self.timezone)
        return to_timezone(out, tzinfo)

    def __eq__(self, other):
        attrs = [
            'hours',
            'minutes',
            'seconds',
            'ordinals',
            'weekdays',
            'months',
            'monthdays',
            'timezone',
        ]
        return all(
            getattr(other, attr, None) == getattr(self, attr, None)
            for attr in attrs
        )

    def __ne__(self, other):
        return not self == other



import datetime
from unittest.mock import MagicMock

import mock
import pytest
import pytz

from testifycompat import assert_equal
from testifycompat import assert_in
from testifycompat import setup
from testifycompat import TestCase
from tests.assertions import assert_call
from tests.assertions import assert_length
from tests.assertions import assert_raises
from tests.testingutils import autospec_method
from tron import actioncommand
from tron import node
from tron.core import action
from tron.core import actiongraph
from tron.core import actionrun
from tron.core import job
from tron.core import jobrun
from tron.serialize import filehandler


def build_mock_job():
    action_graph = mock.create_autospec(actiongraph.ActionGraph)
    action_graph.action_map = {
        'foo': mock.Mock(
            triggered_by=[],
            trigger_timeout=datetime.timedelta(days=1)
        )
    }
    runner = mock.create_autospec(actioncommand.SubprocessActionRunnerFactory)
    return mock.create_autospec(
        job.Job,
        action_graph=action_graph,
        output_path=mock.Mock(),
        context=mock.Mock(),
        action_runner=runner,
    )


class TestJobRun(TestCase):

    now = datetime.datetime(2012, 3, 14, 15, 9, 20, tzinfo=None)
    now_with_tz = datetime.datetime(2012, 3, 14, 15, 9, 20, tzinfo=pytz.utc)

    @setup
    def setup_jobrun(self):
        self.job = build_mock_job()
        self.action_graph = self.job.action_graph
        self.run_time = datetime.datetime(2012, 3, 14, 15, 9, 26)
        mock_node = mock.create_autospec(node.Node)
        self.job_run = jobrun.JobRun(
            'jobname',
            7,
            self.run_time,
            mock_node,
            action_runs=MagicMock(
                action_runs_with_cleanup=[],
                get_startable_action_runs=lambda: [],
            ),
        )
        autospec_method(self.job_run.watch)
        autospec_method(self.job_run.notify)
        self.action_run = mock.create_autospec(
            actionrun.ActionRun,
            is_skipped=False,
        )

    def test__init__(self):
        assert_equal(self.job_run.job_name, 'jobname')
        assert_equal(self.job_run.run_time, self.run_time)
        assert str(self.job_run.output_path).endswith(str(self.job_run.run_num))

    def test_for_job(self):
        run_num = 6
        mock_node = mock.create_autospec(node.Node)
        run = jobrun.JobRun.for_job(
            self.job,
            run_num,
            self.run_time,
            mock_node,
            False,
        )

        assert_equal(run.action_runs.action_graph, self.action_graph)
        assert_equal(run.job_name, self.job.get_name.return_value)
        assert_equal(run.run_num, run_num)
        assert_equal(run.node, mock_node)
        assert not run.manual

    def test_for_job_manual(self):
        run_num = 6
        mock_node = mock.create_autospec(node.Node)
        run = jobrun.JobRun.for_job(
            self.job,
            run_num,
            self.run_time,
            mock_node,
            True,
        )
        assert_equal(run.action_runs.action_graph, self.action_graph)
        assert run.manual

    def test_state_data(self):
        state_data = self.job_run.state_data
        assert_equal(state_data['run_num'], 7)
        assert not state_data['manual']
        assert_equal(state_data['run_time'], self.run_time)

    def test_set_action_runs(self):
        self.job_run._action_runs = None
        count = 2
        action_runs = [
            mock.create_autospec(actionrun.ActionRun) for _ in range(count)
        ]
        run_collection = mock.create_autospec(
            actionrun.ActionRunCollection,
            action_runs_with_cleanup=action_runs,
        )
        self.job_run._set_action_runs(run_collection)
        assert_equal(self.job_run.watch.call_count, count)

        expected = [mock.call(run) for run in action_runs]
        assert_equal(self.job_run.watch.mock_calls, expected)
        assert_equal(self.job_run.action_runs, run_collection)
        assert self.job_run.action_runs_proxy

    def test_set_action_runs_none(self):
        self.job_run._action_runs = None
        run_collection = mock.create_autospec(actionrun.ActionRunCollection)
        self.job_run._set_action_runs(run_collection)
        assert not self.job_run.watch.mock_calls
        assert_equal(self.job_run.action_runs, run_collection)

    def test_set_action_runs_duplicate(self):
        run_collection = mock.create_autospec(actionrun.ActionRunCollection)
        assert_raises(
            ValueError,
            self.job_run._set_action_runs,
            run_collection,
        )

    @mock.patch('tron.core.jobrun.timeutils.current_time', autospec=True)
    def test_seconds_until_run_time(self, mock_current_time):
        mock_current_time.return_value = self.now
        seconds = self.job_run.seconds_until_run_time()
        assert_equal(seconds, 6)

    @mock.patch('tron.core.jobrun.timeutils.current_time', autospec=True)
    def test_seconds_until_run_time_with_tz(self, mock_current_time):
        mock_current_time.return_value = self.now_with_tz
        self.job_run.run_time = self.run_time.replace(tzinfo=pytz.utc)
        seconds = self.job_run.seconds_until_run_time()
        assert_equal(seconds, 6)

    def test_start(self):
        autospec_method(self.job_run._do_start)
        assert self.job_run.start()
        self.job_run._do_start.assert_called_with()

    def test_start_failed(self):
        autospec_method(self.job_run._do_start, return_value=False)
        assert not self.job_run.start()

    def test_do_start(self):
        startable_runs = [
            mock.create_autospec(actionrun.ActionRun) for _ in range(3)
        ]
        self.job_run.action_runs.get_startable_action_runs = lambda: startable_runs

        assert self.job_run._do_start()
        self.job_run.action_runs.ready.assert_called_with()
        for startable_run in startable_runs:
            startable_run.start.assert_called_with()

    def test_do_start_all_failed(self):
        autospec_method(self.job_run._start_action_runs, return_value=[None])
        assert not self.job_run._do_start()

    def test_do_start_some_failed(self):
        returns = [True, None]
        autospec_method(self.job_run._start_action_runs, return_value=returns)
        assert self.job_run._do_start()

    def test_do_start_no_runs(self):
        assert not self.job_run._do_start()

    def test_start_action_runs(self):
        startable_runs = [
            mock.create_autospec(actionrun.ActionRun) for _ in range(3)
        ]
        self.job_run.action_runs.get_startable_action_runs = lambda: startable_runs

        started_runs = self.job_run._start_action_runs()
        assert_equal(started_runs, startable_runs)

    def test_start_action_runs_failed(self):
        startable_runs = [
            mock.create_autospec(actionrun.ActionRun) for _ in range(3)
        ]
        startable_runs[0].start.return_value = False
        self.job_run.action_runs.get_startable_action_runs = lambda: startable_runs

        started_runs = self.job_run._start_action_runs()
        assert_equal(started_runs, startable_runs[1:])

    def test_start_action_runs_all_failed(self):
        startable_runs = [
            mock.create_autospec(actionrun.ActionRun) for _ in range(2)
        ]
        for startable_run in startable_runs:
            startable_run.start.return_value = False
        self.job_run.action_runs.get_startable_action_runs = lambda: startable_runs

        started_runs = self.job_run._start_action_runs()
        assert_equal(started_runs, [])

    def test_handler_trigger_ready_still_scheduled(self):
        autospec_method(self.job_run._start_action_runs)
        self.job_run.is_scheduled = True
        self.job_run.handler(self.action_run, actionrun.ActionRun.NOTIFY_TRIGGER_READY)
        assert not self.job_run._start_action_runs.mock_calls

    def test_handler_trigger_ready_started(self):
        autospec_method(self.job_run._start_action_runs)
        self.job_run.is_scheduled = False
        self.job_run.is_queued = False
        self.job_run.handler(self.action_run, actionrun.ActionRun.NOTIFY_TRIGGER_READY)
        assert self.job_run._start_action_runs.call_count == 1

    def test_handler_not_end_state_event(self):
        autospec_method(self.job_run.finalize)
        autospec_method(self.job_run._start_action_runs)
        self.action_run.is_done = False
        self.job_run.handler(self.action_run, mock.Mock())
        assert not self.job_run.finalize.mock_calls
        assert not self.job_run._start_action_runs.mock_calls

    def test_handler_with_startable(self):
        startable_run = mock.create_autospec(actionrun.ActionRun)
        self.job_run.action_runs.get_startable_action_runs = lambda: [
            startable_run, ]
        autospec_method(self.job_run.finalize)
        self.action_run.is_broken = False

        self.job_run.handler(self.action_run, mock.Mock())
        self.job_run.notify.assert_called_with(
            self.job_run.NOTIFY_STATE_CHANGED,
        )
        startable_run.start.assert_called_with()
        assert not self.job_run.finalize.mock_calls

    def test_handler_runs_not_done(self):
        self.job_run.action_runs.is_done = False
        autospec_method(self.job_run._start_action_runs, return_value=[])
        autospec_method(self.job_run.finalize)
        self.job_run.handler(self.action_run, mock.Mock())
        assert not self.job_run.finalize.mock_calls

    def test_handler_finished_without_cleanup(self):
        self.job_run.action_runs.is_active = False
        self.job_run.action_runs.is_scheduled = False
        self.job_run.action_runs.cleanup_action_run = None
        autospec_method(self.job_run.finalize)
        self.job_run.handler(self.action_run, mock.Mock())
        self.job_run.finalize.assert_called_with()

    def test_handler_finished_with_cleanup_done(self):
        self.job_run.action_runs.is_active = False
        self.job_run.action_runs.is_scheduled = False
        self.job_run.action_runs.cleanup_action_run = mock.Mock(is_done=True)
        autospec_method(self.job_run.finalize)
        self.job_run.handler(self.action_run, mock.Mock())
        self.job_run.finalize.assert_called_with()

    def test_handler_finished_with_cleanup(self):
        self.job_run.action_runs.is_active = False
        self.job_run.action_runs.is_scheduled = False
        self.job_run.action_runs.cleanup_action_run = mock.Mock(is_done=False)
        autospec_method(self.job_run.finalize)
        self.job_run.handler(self.action_run, mock.Mock())
        assert not self.job_run.finalize.mock_calls
        self.job_run.action_runs.cleanup_action_run.start.assert_called_with()

    def test_handler_action_run_cancelled(self):
        self.action_run.is_broken = True
        autospec_method(self.job_run._start_action_runs)
        self.job_run.handler(self.action_run, mock.Mock())
        assert not self.job_run._start_action_runs.mock_calls

    def test_handler_action_run_skipped(self):
        self.action_run.is_broken = False
        self.action_run.is_skipped = True
        self.job_run.action_runs.is_scheduled = True
        autospec_method(self.job_run._start_action_runs)
        self.job_run.handler(self.action_run, mock.Mock())
        assert not self.job_run._start_action_runs.mock_calls

    def test_state(self):
        assert_equal(self.job_run.state, actionrun.ActionRun.SUCCEEDED)

    def test_state_with_no_action_runs(self):
        self.job_run._action_runs = None
        assert_equal(self.job_run.state, actionrun.ActionRun.UNKNOWN)

    def test_finalize(self):
        self.job_run.action_runs.is_failed = False
        self.job_run.finalize()
        self.job_run.notify.assert_called_with(self.job_run.NOTIFY_DONE)

    def test_finalize_failure(self):
        self.job_run.finalize()
        self.job_run.notify.assert_called_with(self.job_run.NOTIFY_DONE)

    def test_cleanup(self):
        autospec_method(self.job_run.clear_observers)
        self.job_run.output_path = mock.create_autospec(filehandler.OutputPath)
        self.job_run.cleanup()

        self.job_run.clear_observers.assert_called_with()
        self.job_run.output_path.delete.assert_called_with()
        assert not self.job_run.node
        assert not self.job_run.action_graph
        assert not self.job_run.action_runs

    def test__getattr__(self):
        assert self.job_run.cancel
        assert self.job_run.state == 'succeeded'
        assert self.job_run.is_succeeded

    def test__getattr__miss(self):
        assert_raises(AttributeError, lambda: self.job_run.bogus)


class TestJobRunFromState(TestCase):
    @setup
    def setup_jobrun(self):
        self.action_graph = mock.create_autospec(actiongraph.ActionGraph)
        self.run_time = datetime.datetime(2012, 3, 14, 15, 9, 26)
        self.path = ['base', 'path']
        self.output_path = mock.create_autospec(filehandler.OutputPath)
        self.node_pool = mock.create_autospec(node.NodePool)
        self.action_run_state_data = [{
            'job_run_id': 'thejobname.22',
            'action_name': 'blingaction',
            'state': 'succeeded',
            'run_time': 'sometime',
            'start_time': 'sometime',
            'end_time': 'sometime',
            'command': 'doit',
            'node_name': 'thenode',
        }]
        self.state_data = {
            'job_name': 'thejobname',
            'run_num': 22,
            'run_time': self.run_time,
            'node_name': 'thebox',
            'end_time': 'the_end',
            'start_time': 'start_time',
            'runs': self.action_run_state_data,
            'cleanup_run': None,
            'manual': True,
        }
        self.context = mock.Mock()

    def test_from_state(self):
        run = jobrun.JobRun.from_state(
            self.state_data,
            self.action_graph,
            self.output_path,
            self.context,
            self.node_pool,
        )
        assert_length(run.action_runs.run_map, 1)
        assert_equal(run.job_name, self.state_data['job_name'])
        assert_equal(run.run_time, self.run_time)
        assert run.manual
        assert_equal(run.output_path, self.output_path)
        assert run.context.next
        assert run.action_graph

    def test_from_state_node_no_longer_exists(self):
        run = jobrun.JobRun.from_state(
            self.state_data,
            self.action_graph,
            self.output_path,
            self.context,
            self.node_pool,
        )
        assert_length(run.action_runs.run_map, 1)
        assert_equal(run.job_name, 'thejobname')
        assert_equal(run.run_time, self.run_time)
        assert_equal(run.node, self.node_pool)


class MockJobRun(MagicMock):

    manual = False

    node = 'anode'

    @property
    def is_scheduled(self):
        return self.state == actionrun.ActionRun.SCHEDULED

    @property
    def is_queued(self):
        return self.state == actionrun.ActionRun.QUEUED

    @property
    def is_running(self):
        return self.state == actionrun.ActionRun.RUNNING

    @property
    def is_starting(self):
        return self.state == actionrun.ActionRun.STARTING

    @property
    def is_waiting(self):
        return self.state == actionrun.ActionRun.WAITING

    def __repr__(self):
        return str(self.__dict__)


class TestJobRunCollection(TestCase):
    def _mock_run(self, **kwargs):
        return MockJobRun(**kwargs)

    @setup
    def setup_runs(self):
        self.run_collection = jobrun.JobRunCollection(6)
        self.job_runs = [
            self._mock_run(state=actionrun.ActionRun.QUEUED, run_num=5),
            self._mock_run(state=actionrun.ActionRun.WAITING, run_num=4),
            self._mock_run(state=actionrun.ActionRun.RUNNING, run_num=3),
        ] + [
            self._mock_run(
                state=actionrun.ActionRun.SUCCEEDED,
                run_num=i,
            ) for i in range(2, 0, -1)
        ]
        self.run_collection.runs.extend(self.job_runs)
        self.mock_node = mock.create_autospec(node.Node)

    def test__init__(self):
        assert_equal(self.run_collection.run_limit, 6)

    def test_from_config(self):
        job_config = mock.Mock(run_limit=20)
        runs = jobrun.JobRunCollection.from_config(job_config)
        assert_equal(runs.run_limit, 20)

    def test_job_runs_from_state(self):
        state_data = [
            dict(
                run_num=i,
                job_name="thename",
                run_time="sometime",
                start_time="start_time",
                end_time="sometime",
                cleanup_run=None,
                runs=[],
            ) for i in range(3, -1, -1)
        ]
        action_graph = mock.create_autospec(actiongraph.ActionGraph)
        output_path = mock.create_autospec(filehandler.OutputPath)
        context = mock.Mock()
        node_pool = mock.create_autospec(node.NodePool)
        runs = jobrun.job_runs_from_state(
            state_data,
            action_graph,
            output_path,
            context,
            node_pool,
        )
        assert len(runs) == 4
        assert all([type(job) == jobrun.JobRun for job in runs])

    def test_build_new_run(self):
        autospec_method(self.run_collection.remove_old_runs)
        run_time = datetime.datetime(2012, 3, 14, 15, 9, 26)
        mock_job = build_mock_job()
        job_run = self.run_collection.build_new_run(
            mock_job,
            run_time,
            self.mock_node,
        )
        assert_in(job_run, self.run_collection.runs)
        self.run_collection.remove_old_runs.assert_called_with()
        assert job_run.run_num == 6
        assert job_run.job_name == mock_job.get_name.return_value

    def test_build_new_run_manual(self):
        autospec_method(self.run_collection.remove_old_runs)
        run_time = datetime.datetime(2012, 3, 14, 15, 9, 26)
        mock_job = build_mock_job()
        job_run = self.run_collection.build_new_run(
            mock_job,
            run_time,
            self.mock_node,
            True,
        )
        assert_in(job_run, self.run_collection.runs)
        self.run_collection.remove_old_runs.assert_called_with()
        assert job_run.run_num == 6
        assert job_run.manual

    def test_cancel_pending(self):
        pending_runs = [mock.Mock() for _ in range(2)]
        autospec_method(
            self.run_collection.get_pending,
            return_value=pending_runs,
        )
        self.run_collection.cancel_pending()
        for pending_run in pending_runs:
            pending_run.cancel.assert_called_with()

    def test_cancel_pending_no_pending(self):
        autospec_method(self.run_collection.get_pending, return_value=[])
        self.run_collection.cancel_pending()

    def test_remove_pending(self):
        self.run_collection.remove_pending()
        assert_length(self.run_collection.runs, 4)
        assert_equal(self.run_collection.runs[0], self.job_runs[1])
        assert_call(self.job_runs[0].cleanup, 0)

    def test_get_run_by_state(self):
        state = actionrun.ActionRun.SUCCEEDED
        run = self.run_collection.get_run_by_state(state)
        assert_equal(run, self.job_runs[3])

    def test_get_run_by_state_no_match(self):
        state = actionrun.ActionRun.UNKNOWN
        run = self.run_collection.get_run_by_state(state)
        assert_equal(run, None)

    def test_get_run_by_num(self):
        run = self.run_collection.get_run_by_num(1)
        assert_equal(run.run_num, 1)

    def test_get_run_by_num_no_match(self):
        run = self.run_collection.get_run_by_num(7)
        assert_equal(run, None)

    def test_get_run_by_index(self):
        run = self.run_collection.get_run_by_index(-1)
        assert_equal(run, self.job_runs[0])
        run = self.run_collection.get_run_by_index(-2)
        assert_equal(run, self.job_runs[1])
        run = self.run_collection.get_run_by_index(0)
        assert_equal(run, self.job_runs[-1])
        run = self.run_collection.get_run_by_index(1)
        assert_equal(run, self.job_runs[-2])

    def test_get_run_by_index_invalid_index(self):
        run = self.run_collection.get_run_by_index(-6)
        assert_equal(run, None)
        run = self.run_collection.get_run_by_index(5)
        assert_equal(run, None)

    def test_get_newest(self):
        run = self.run_collection.get_newest()
        assert_equal(run, self.job_runs[0])

    def test_get_newest_exclude_manual(self):
        run = self._mock_run(
            state=actionrun.ActionRun.RUNNING,
            run_num=5,
            manual=True,
        )
        self.job_runs.insert(0, run)
        newest_run = self.run_collection.get_newest(include_manual=False)
        assert_equal(newest_run, self.job_runs[1])

    def test_get_newest_no_runs(self):
        run_collection = jobrun.JobRunCollection(5)
        assert_equal(run_collection.get_newest(), None)

    def test_pending(self):
        run_num = self.run_collection.next_run_num()
        scheduled_run = self._mock_run(
            run_num=run_num,
            state=actionrun.ActionRun.SCHEDULED,
        )
        self.run_collection.runs.appendleft(scheduled_run)
        pending = list(self.run_collection.get_pending())
        assert_length(pending, 2)
        assert_equal(pending, [scheduled_run, self.job_runs[0]])

    def test_get_active(self):
        starting_run = self._mock_run(
            run_num=self.run_collection.next_run_num(),
            state=actionrun.ActionRun.STARTING,
        )
        self.run_collection.runs.appendleft(starting_run)
        active = list(self.run_collection.get_active())
        assert_length(active, 3)
        assert_equal(active, [starting_run, self.job_runs[1], self.job_runs[2]])

    def test_get_active_with_node(self):
        starting_run = self._mock_run(
            run_num=self.run_collection.next_run_num(),
            state=actionrun.ActionRun.STARTING,
        )
        starting_run.node = 'differentnode'
        self.run_collection.runs.appendleft(starting_run)
        active = list(self.run_collection.get_active('anode'))
        assert_length(active, 2)
        assert_equal(active, [self.job_runs[1], self.job_runs[2]])

    def test_get_active_none(self):
        active = list(self.run_collection.get_active('bogus'))
        assert_length(active, 0)

    def test_get_first_queued(self):
        run_num = self.run_collection.next_run_num()
        second_queued = self._mock_run(
            run_num=run_num,
            state=actionrun.ActionRun.QUEUED,
        )
        self.run_collection.runs.appendleft(second_queued)

        first_queued = self.run_collection.get_first_queued()
        assert_equal(first_queued, self.job_runs[0])

    def test_get_first_queued_no_match(self):
        self.job_runs[0].state = actionrun.ActionRun.CANCELLED
        first_queued = self.run_collection.get_first_queued()
        assert not first_queued

    def test_get_next_run_num(self):
        assert_equal(self.run_collection.next_run_num(), 6)

    def test_get_next_run_num_first(self):
        run_collection = jobrun.JobRunCollection(5)
        assert_equal(run_collection.next_run_num(), 0)

    def test_remove_old_runs(self):
        self.run_collection.run_limit = 1
        self.run_collection.remove_old_runs()

        assert_length(self.run_collection.runs, 1)
        assert_call(self.job_runs[-1].cleanup, 0)
        for job_run in self.run_collection.runs:
            assert_length(job_run.cancel.calls, 0)

    def test_remove_old_runs_none(self):
        self.run_collection.remove_old_runs()
        for job_run in self.job_runs:
            assert_length(job_run.cancel.calls, 0)

    def test_remove_old_runs_no_runs(self):
        run_collection = jobrun.JobRunCollection(4)
        run_collection.remove_old_runs()

    def test_state_data(self):
        assert_length(self.run_collection.state_data, len(self.job_runs))

    def test_last_success(self):
        assert_equal(self.run_collection.last_success, self.job_runs[3])

    def test__str__(self):
        expected = "JobRunCollection[5(queued), 4(waiting), 3(running), 2(succeeded), 1(succeeded)]"
        assert_equal(str(self.run_collection), expected)

    def test_get_action_runs(self):
        action_name = 'action_name'
        self.run_collection.runs = job_runs = [mock.Mock(), mock.Mock()]
        runs = self.run_collection.get_action_runs(action_name)
        expected = [
            job_run.get_action_run.return_value for job_run in job_runs
        ]
        assert_equal(runs, expected)
        for job_run in job_runs:
            job_run.get_action_run.assert_called_with(action_name)


class TestJobRunStateTransitions:
    """Integration test for the state of a job run when actions change state in various ways."""

    @pytest.fixture
    def mock_event_bus(self):
        with mock.patch(
            'tron.core.actionrun.EventBus',
            autospec=True,
        ) as mock_event_bus:
            mock_event_bus.has_event.return_value = True
            yield mock_event_bus

    @pytest.fixture
    def job_run(self, tmpdir, mock_event_bus):
        action_foo = action.Action('foo', 'command', None)
        action_after_foo = action.Action('after_foo', 'command', None)
        action_bar = action.Action('bar', 'command', None, triggered_by={'trigger'})
        action_graph = actiongraph.ActionGraph(
            action_map={
                'foo': action_foo,
                'after_foo': action_after_foo,
                'bar': action_bar,
            },
            required_actions={'foo': set(), 'after_foo': {'foo'}, 'bar': set()},
            required_triggers={'foo': set(), 'after_foo': set(), 'bar': {'trigger'}},
        )
        mock_job = mock.Mock(
            output_path=filehandler.OutputPath(tmpdir),
            action_graph=action_graph,
            action_runner=actioncommand.NoActionRunnerFactory(),
        )
        job_run = jobrun.JobRun.for_job(
            mock_job,
            run_num=1,
            run_time=datetime.datetime.now(),
            node=mock.Mock(),
            manual=False,
        )
        return job_run

    def test_success_path(self, job_run):
        # Check expected states as actions run normally and succeed.
        foo = job_run.get_action_run('foo')
        after_foo = job_run.get_action_run('after_foo')
        bar = job_run.get_action_run('bar')

        # Run is initially SCHEDULED
        assert job_run.state == actionrun.ActionRun.SCHEDULED

        # After starting, both actions without dependencies start.
        # Run is STARTING
        job_run.start()
        assert foo.is_starting
        assert bar.is_starting
        assert job_run.state == actionrun.ActionRun.STARTING

        # Commands start successfully, run is RUNNING.
        foo.action_command.started()
        bar.action_command.started()
        assert job_run.state == actionrun.ActionRun.RUNNING

        # Still RUNNING after one of two running actions succeeds
        bar.action_command.exited(0)
        assert job_run.state == actionrun.ActionRun.RUNNING

        # after_foo starts after its dependency succeeds
        foo.action_command.exited(0)
        assert after_foo.is_starting
        after_foo.action_command.started()
        assert job_run.state == actionrun.ActionRun.RUNNING

        # SUCCEEDED after all actions succeed
        after_foo.action_command.exited(0)
        assert job_run.state == actionrun.ActionRun.SUCCEEDED

    def test_one_action_fails(self, job_run):
        foo = job_run.get_action_run('foo')
        after_foo = job_run.get_action_run('after_foo')
        bar = job_run.get_action_run('bar')

        # bar action fails, job is RUNNING because foo is still running
        job_run.start()
        foo.action_command.started()
        bar.action_command.started()
        bar.action_command.exited(1)
        assert job_run.state == actionrun.ActionRun.RUNNING

        # after_foo still starts after its dependency succeeds
        foo.action_command.exited(0)
        assert after_foo.is_starting
        after_foo.action_command.started()
        assert job_run.state == actionrun.ActionRun.RUNNING

        # After running actions finish, run enters FAILED terminal state
        after_foo.action_command.exited(0)
        assert job_run.state == actionrun.ActionRun.FAILED

        # If we skip the failed action, run becomes SUCCEEDED
        bar.skip()
        assert job_run.state == actionrun.ActionRun.SUCCEEDED

    def test_one_action_unknown(self, job_run):
        foo = job_run.get_action_run('foo')
        after_foo = job_run.get_action_run('after_foo')
        bar = job_run.get_action_run('bar')

        assert job_run.state == actionrun.ActionRun.SCHEDULED

        # bar action becomes unknown, job is RUNNING because foo is still running
        job_run.start()
        foo.action_command.started()
        bar.action_command.started()
        bar.action_command.exited(None)
        assert job_run.state == actionrun.ActionRun.RUNNING

        # after_foo still starts after its dependency succeeds
        foo.action_command.exited(0)
        assert after_foo.is_starting
        after_foo.action_command.started()
        assert job_run.state == actionrun.ActionRun.RUNNING

        # UNKNOWN after running actions finish
        after_foo.action_command.exited(0)
        assert job_run.state == actionrun.ActionRun.UNKNOWN

    def test_both_unknown_and_failed(self, job_run):
        foo = job_run.get_action_run('foo')
        after_foo = job_run.get_action_run('after_foo')
        bar = job_run.get_action_run('bar')

        # bar action becomes unknown, job is RUNNING because foo is still running
        job_run.start()
        foo.action_command.started()
        bar.action_command.started()
        bar.action_command.exited(None)
        assert job_run.state == actionrun.ActionRun.RUNNING

        # after_foo still starts after its dependency succeeds
        foo.action_command.exited(0)
        assert after_foo.is_starting
        after_foo.action_command.started()
        assert job_run.state == actionrun.ActionRun.RUNNING

        # A different action fails
        # Overall run is FAILED
        after_foo.action_command.exited(1)
        assert job_run.state == actionrun.ActionRun.FAILED

    def test_required_action_fails(self, job_run):
        foo = job_run.get_action_run('foo')
        after_foo = job_run.get_action_run('after_foo')
        bar = job_run.get_action_run('bar')

        assert job_run.state == actionrun.ActionRun.SCHEDULED

        # An action (foo) required by another action fails
        # Run is RUNNING while the other action, bar, is running
        job_run.start()
        foo.action_command.started()
        bar.action_command.started()
        foo.action_command.exited(1)
        assert job_run.state == actionrun.ActionRun.RUNNING

        # bar action succeeds
        # after_foo cannot run because its required action failed
        # So run is FAILED even though after_foo is waiting
        bar.action_command.exited(0)
        assert after_foo.is_waiting
        assert job_run.state == actionrun.ActionRun.FAILED

        # Pretend we reconfigured and after_foo doesn't depend on foo anymore
        # Run should not be WAITING
        # Ideally it would still be FAILED, but for now it's UNKNOWN in this case
        job_run.action_runs.action_graph.required_actions['after_foo'] = {}
        assert job_run.state == actionrun.ActionRun.UNKNOWN

    def test_required_action_unknown(self, job_run):
        foo = job_run.get_action_run('foo')
        after_foo = job_run.get_action_run('after_foo')
        bar = job_run.get_action_run('bar')

        # An action (foo) required by another action becomes unknown
        # Run is RUNNING while the other action, bar, is running
        job_run.start()
        foo.action_command.started()
        bar.action_command.started()
        foo.action_command.exited(None)
        assert job_run.state == actionrun.ActionRun.RUNNING

        # Other action succeeds
        # after_foo cannot run because its required action is unknown
        # So run is UNKNOWN even though after_foo is waiting
        bar.action_command.exited(0)
        assert after_foo.is_waiting
        assert job_run.state == actionrun.ActionRun.UNKNOWN

        # Pretend we reconfigured and after_foo doesn't depend on foo anymore
        # Run should not be waiting
        job_run.action_runs.action_graph.required_actions['after_foo'] = {}
        assert job_run.state == actionrun.ActionRun.UNKNOWN

    def test_with_trigger(self, job_run, mock_event_bus):
        foo = job_run.get_action_run('foo')
        after_foo = job_run.get_action_run('after_foo')
        bar = job_run.get_action_run('bar')

        # Start without trigger for bar
        mock_event_bus.has_event.return_value = False
        # Job should still start in scheduled state
        assert job_run.state == actionrun.ActionRun.SCHEDULED

        # Only foo is able to start
        job_run.start()
        assert foo.is_starting
        assert bar.is_waiting
        assert job_run.state == actionrun.ActionRun.STARTING

        foo.action_command.started()
        assert job_run.state == actionrun.ActionRun.RUNNING

        # after_foo runs normally after foo succeeds
        foo.action_command.exited(0)
        assert after_foo.is_starting
        after_foo.action_command.started()
        assert job_run.state == actionrun.ActionRun.RUNNING

        # After after_foo succeeds, run is not done
        # WAITING because bar is still waiting for a trigger
        after_foo.action_command.exited(0)
        assert job_run.state == actionrun.ActionRun.WAITING

        # After trigger is available, job run finishes as normal
        mock_event_bus.has_event.return_value = True
        bar.trigger_notify()
        assert bar.is_starting
        bar.action_command.started()
        bar.action_command.exited(0)
        assert job_run.state == actionrun.ActionRun.SUCCEEDED

    def test_queued(self, job_run):
        assert job_run.state == actionrun.ActionRun.SCHEDULED
        job_run.queue()
        assert job_run.state == actionrun.ActionRun.QUEUED
        job_run.start()
        assert job_run.state == actionrun.ActionRun.STARTING

    def test_cancel_one(self, job_run):
        assert job_run.state == actionrun.ActionRun.SCHEDULED
        job_run.start()
        assert job_run.state == actionrun.ActionRun.STARTING
        job_run.get_action_run('after_foo').cancel()
        assert job_run.state == actionrun.ActionRun.CANCELLED



from __future__ import absolute_import
from __future__ import unicode_literals

import tempfile

import mock
import pytest

from testifycompat import assert_equal
from testifycompat import setup
from testifycompat import setup_teardown
from testifycompat import TestCase
from tron.bin import action_runner


class TestStatusFile(TestCase):
    @setup
    def setup_status_file(self):
        self.filename = tempfile.NamedTemporaryFile().name
        self.status_file = action_runner.StatusFile(self.filename)

    def test_get_content(self):
        command, proc, run_id = 'do this', mock.Mock(), 'Job.test.1'
        with mock.patch('tron.bin.action_runner.time.time', autospec=True) as faketime, \
                mock.patch('tron.bin.action_runner.os.getpid', autospec=True) as fakepid:
            faketime.return_value = 0
            fakepid.return_value = 2
            content = self.status_file.get_content(
                command=command,
                proc=proc,
                run_id=run_id,
            )
            expected = dict(
                run_id=run_id,
                command=command,
                pid=proc.pid,
                return_code=proc.returncode,
                runner_pid=2,
                timestamp=0,
            )
        assert_equal(content, expected)


class TestRegister(TestCase):
    mock_isdir = mock_status_file = None
    mock_makedirs = None

    @setup_teardown
    def patch_sys(self):
        with mock.patch('tron.bin.action_runner.os.path.isdir', autospec=True) as self.mock_isdir, \
                mock.patch('tron.bin.action_runner.os.makedirs', autospec=True) as self.mock_makedirs, \
                mock.patch('tron.bin.action_runner.os.access', autospec=True) as self.mock_access, \
                mock.patch('tron.bin.action_runner.StatusFile', autospec=True) as self.mock_status_file:
            self.output_path = '/bogus/path/does/not/exist'
            self.command = 'command'
            self.run_id = 'Job.test.1'
            self.proc = mock.Mock()
            self.proc.wait.return_value = 0
            yield

    def test_validate_output_dir_does_not_exist(self):
        self.mock_isdir.return_value = False
        self.mock_access.return_value = True
        action_runner.validate_output_dir(self.output_path)
        self.mock_makedirs.assert_called_with(self.output_path)

    def test_validate_output_dir_does_not_exist_create_fails(self):
        self.mock_isdir.return_value = False
        self.mock_access.return_value = True
        self.mock_makedirs.side_effect = OSError
        with pytest.raises(OSError):
            action_runner.validate_output_dir(self.output_path)

    def test_validate_output_dir_exists_not_writable(self):
        self.mock_isdir.return_value = True
        self.mock_access.return_value = False
        with pytest.raises(OSError):
            action_runner.validate_output_dir(self.output_path)

    def test_run_proc(self):
        self.mock_isdir.return_value = True
        self.mock_access.return_value = True
        action_runner.run_proc(
            self.output_path,
            self.command,
            self.run_id,
            self.proc,
        )
        self.mock_status_file.assert_called_with(
            self.output_path + '/' + action_runner.STATUS_FILE,
        )
        self.mock_status_file.return_value.wrap.assert_called_with(
            command=self.command,
            run_id=self.run_id,
            proc=self.proc,
        )
        self.proc.wait.assert_called_with()


class TestBuildEnvironment:
    def test_build_environment(self):
        with mock.patch('tron.bin.action_runner.os.environ', dict(PATH='/usr/bin/nowhere'), autospec=None):
            env = action_runner.build_environment('MASTER.foo.10.bar')

        assert env == dict(
            PATH='/usr/bin/nowhere',
            TRON_JOB_NAMESPACE='MASTER',
            TRON_JOB_NAME='foo',
            TRON_RUN_NUM='10',
            TRON_ACTION='bar',
        )

    def test_build_environment_invalid_run_id(self):
        with mock.patch('tron.bin.action_runner.os.environ', dict(PATH='/usr/bin/nowhere'), autospec=None):
            env = action_runner.build_environment('asdf')

        assert env == dict(
            PATH='/usr/bin/nowhere',
            TRON_JOB_NAMESPACE='UNKNOWN',
            TRON_JOB_NAME='UNKNOWN',
            TRON_RUN_NUM='UNKNOWN',
            TRON_ACTION='UNKNOWN',
        )

    def test_build_environment_too_long_run_id(self):
        with mock.patch('tron.bin.action_runner.os.environ', dict(PATH='/usr/bin/nowhere'), autospec=None):
            env = action_runner.build_environment('MASTER.foo.10.bar.baz')

        assert env == dict(
            PATH='/usr/bin/nowhere',
            TRON_JOB_NAMESPACE='MASTER',
            TRON_JOB_NAME='foo',
            TRON_RUN_NUM='10',
            TRON_ACTION='bar.baz',
        )



from unittest import mock

import pytest

from tron.config.schema import ConfigAction
from tron.config.schema import ConfigJob
from tron.core.jobgraph import AdjListEntry
from tron.core.jobgraph import JobGraph


MISSING_DEPENDENCY_ERR_MSG = '''The following actions are dependencies of other actions but missing:
Action other.job2.action3 is dependency of actions:
  - MASTER.job3.action5
Please check if you have deleted/renamed any of them or their containing jobs.'''


def _setup_job_graph_config_container():
    action1 = ConfigAction(
        name='action1',
        command='do something',
    )
    action2 = ConfigAction(
        name='action2',
        command='do something',
        requires=['action1'],
    )
    job1_config = ConfigJob(
        name='job1',
        node='default',
        schedule=mock.Mock(),
        actions={'action1': action1, 'action2': action2},
        namespace='MASTER',
    )

    action3 = ConfigAction(
        name='action3',
        command='do something',
        triggered_by=['MASTER.job1.action2.shortdate.{shortdate}'],
    )
    job2_config = ConfigJob(
        name='job1',
        node='default',
        schedule=mock.Mock(),
        actions={'action3': action3},
        namespace='other',
    )

    action4 = ConfigAction(
        name='action4',
        command='do something',
    )
    action5 = ConfigAction(
        name='action5',
        command='do something',
        requires=['action4'],
        triggered_by=['other.job2.action3.shortdate.{shortdate}'],
    )
    job3_config = ConfigJob(
        name='job1',
        node='default',
        schedule=mock.Mock(),
        actions={'action4': action4, 'action5': action5},
        namespace='MASTER',
    )
    config_container = mock.Mock()
    config_container.get_jobs.return_value = {
        'MASTER.job1': job1_config,
        'other.job2': job2_config,
        'MASTER.job3': job3_config,
    }
    return config_container


class TestJobGraph:

    def setup_method(self):
        self.job_graph = JobGraph(_setup_job_graph_config_container(), should_validate_missing_dependency=True)

    def test_job_graph_missing_dependency(self):
        missing_dependency_config_container = _setup_job_graph_config_container()
        missing_dependency_config_container.get_jobs.return_value.pop('other.job2')
        with pytest.raises(ValueError) as e:
            JobGraph(missing_dependency_config_container, should_validate_missing_dependency=True)
        assert str(e.value) == MISSING_DEPENDENCY_ERR_MSG

    def test_job_graph(self):
        assert sorted(list(self.job_graph.action_map.keys())) == [
            'MASTER.job1.action1',
            'MASTER.job1.action2',
            'MASTER.job3.action4',
            'MASTER.job3.action5',
            'other.job2.action3',
        ]
        assert self.job_graph._actions_for_job == {
            'MASTER.job1': ['MASTER.job1.action1', 'MASTER.job1.action2'],
            'other.job2': ['other.job2.action3'],
            'MASTER.job3': ['MASTER.job3.action4', 'MASTER.job3.action5'],
        }
        assert self.job_graph._adj_list == {
            'MASTER.job1.action1': [AdjListEntry('MASTER.job1.action2', False)],
            'MASTER.job1.action2': [AdjListEntry('other.job2.action3', True)],
            'other.job2.action3': [AdjListEntry('MASTER.job3.action5', True)],
            'MASTER.job3.action4': [AdjListEntry('MASTER.job3.action5', False)],
        }
        assert self.job_graph._rev_adj_list == {
            'MASTER.job1.action1': [],
            'MASTER.job1.action2': [AdjListEntry('MASTER.job1.action1', False)],
            'other.job2.action3': [AdjListEntry('MASTER.job1.action2', True)],
            'MASTER.job3.action4': [],
            'MASTER.job3.action5': [
                AdjListEntry('MASTER.job3.action4', False),
                AdjListEntry('other.job2.action3', True),
            ],
        }

    def test_get_action_graph_for_job(self):
        action_graph_1 = self.job_graph.get_action_graph_for_job('MASTER.job1')
        assert sorted((action_graph_1.action_map.keys())) == [
            'action1',
            'action2',
        ]
        assert action_graph_1.required_actions == {
            'action1': set(),
            'action2': {'action1'}
        }
        assert action_graph_1.required_triggers == {
            'other.job2.action3': {'action2'},
            'MASTER.job3.action5': {'other.job2.action3'},
        }

        action_graph_2 = self.job_graph.get_action_graph_for_job('other.job2')
        assert sorted((action_graph_2.action_map.keys())) == [
            'action3',
        ]
        assert action_graph_2.required_actions == {
            'action3': set(),
        }
        assert action_graph_2.required_triggers == {
            'action3': {'MASTER.job1.action2'},
            'MASTER.job3.action5': {'action3'},
        }

        action_graph_3 = self.job_graph.get_action_graph_for_job('MASTER.job3')
        assert sorted((action_graph_3.action_map.keys())) == [
            'action4',
            'action5',
        ]
        assert action_graph_3.required_actions == {
            'action4': set(),
            'action5': {'action4'},
        }
        assert action_graph_3.required_triggers == {
            'action5': {'other.job2.action3'},
            'other.job2.action3': {'MASTER.job1.action2'},
        }



import itertools
import logging
import time
from contextlib import contextmanager

from tron.config import schema
from tron.core import job
from tron.mesos import MesosClusterRepository
from tron.serialize import runstate
from tron.serialize.runstate.dynamodb_state_store import DynamoDBStateStore
from tron.serialize.runstate.shelvestore import ShelveStateStore
from tron.serialize.runstate.yamlstore import YamlStateStore
from tron.utils import observer

log = logging.getLogger(__name__)


class VersionMismatchError(ValueError):
    """Raised when the state has a newer version then tron.__version.__."""


class PersistenceStoreError(ValueError):
    """Raised if the store can not be created or fails a read or write."""


class PersistenceManagerFactory(object):
    """Create a PersistentStateManager."""

    @classmethod
    def from_config(cls, persistence_config):
        store_type = schema.StatePersistenceTypes(persistence_config.store_type)
        name = persistence_config.name
        buffer_size = persistence_config.buffer_size
        store = None

        if store_type == schema.StatePersistenceTypes.shelve:
            store = ShelveStateStore(name)

        if store_type == schema.StatePersistenceTypes.yaml:
            store = YamlStateStore(name)

        if store_type == schema.StatePersistenceTypes.dynamodb:
            table_name = persistence_config.table_name
            dynamodb_region = persistence_config.dynamodb_region
            store = DynamoDBStateStore(table_name, dynamodb_region)

        buffer = StateSaveBuffer(buffer_size)
        return PersistentStateManager(store, buffer)


class StateMetadata(object):
    """A data object for saving state metadata. Conforms to the same
    RunState interface as Jobs and Services.
    """
    name = 'StateMetadata'

    # State schema version, only first component counts,
    # for backwards compatibility
    version = (0, 7, 0, 0)

    def __init__(self):
        self.state_data = {
            'version': self.version,
            'create_time': time.time(),
        }

    @classmethod
    def validate_metadata(cls, metadata):
        """Raises an exception if the metadata version is newer then
        StateMetadata.version
        """
        if not metadata:
            return

        if metadata['version'][0] > cls.version[0]:
            msg = "State version %s, expected <= %s"
            raise VersionMismatchError(
                msg % (
                    metadata['version'],
                    cls.version,
                ),
            )


class StateSaveBuffer(object):
    """Buffer calls to save, and perform the saves when buffer reaches
    buffer size. This buffer will only store one state_data for each key.
    """

    def __init__(self, buffer_size):
        self.buffer_size = buffer_size
        self.buffer = {}
        self.counter = itertools.cycle(range(buffer_size))

    def save(self, key, state_data):
        """Save the state_data indexed by key and return True if the buffer
        is full.
        """
        self.buffer[key] = state_data
        return not next(self.counter)

    def __iter__(self):
        """Return all buffered data and clear the buffer."""
        for key, item in self.buffer.items():
            yield key, item
        self.buffer.clear()


class PersistentStateManager(object):
    """Provides an interface to persist the state of Tron.

    The implementation of persisting and restoring the state from disk is
    handled by a class which supports the StateStore interface:

    class IStateStore(object):

        def build_key(self, type, identifier):
            return <a key>

        def restore(self, keys):
            return <dict of key to states>

        def save(self, key, state_data):
            pass

        def cleanup(self):
            pass

    """

    def __init__(self, persistence_impl, buffer):
        self.enabled = True
        self._buffer = buffer
        self._impl = persistence_impl
        self.metadata_key = self._impl.build_key(
            runstate.MCP_STATE,
            StateMetadata.name,
        )

    def restore(self, job_names, skip_validation=False):
        """Return the most recent serialized state."""
        log.debug("Restoring state.")
        if not skip_validation:
            self._restore_metadata()

        jobs = self._restore_dicts(runstate.JOB_STATE, job_names)
        frameworks = self._restore_dicts(runstate.MESOS_STATE, ['frameworks'])

        state = {
            runstate.JOB_STATE: jobs,
            runstate.MESOS_STATE: frameworks,
        }
        return state

    def _restore_metadata(self):
        metadata = self._impl.restore([self.metadata_key])
        StateMetadata.validate_metadata(metadata.get(self.metadata_key))

    def _keys_for_items(self, item_type, names):
        """Returns a dict of item to the key for that item."""
        keys = (self._impl.build_key(item_type, name) for name in names)
        return dict(zip(keys, names))

    def _restore_dicts(self, item_type, items):
        """Return a dict mapping of the items name to its state data."""
        key_to_item_map = self._keys_for_items(item_type, items)
        key_to_state_map = self._impl.restore(key_to_item_map.keys())
        return {
            key_to_item_map[key]: state_data
            for key, state_data in key_to_state_map.items()
        }

    def save(self, type_enum, name, state_data):
        """Persist an items state."""
        key = self._impl.build_key(type_enum, name)
        log.debug("Buffering state save for: %s", key)
        if self._buffer.save(key, state_data):
            if not self.enabled:
                log.debug(f"State manager disabled, not persisting {key}")
                return
            self._save_from_buffer()

    def _save_from_buffer(self):
        key_state_pairs = list(self._buffer)
        if not key_state_pairs:
            return

        keys = ','.join(str(key) for key, _ in key_state_pairs)
        log.info("Saving state for %s" % keys)

        with self._timeit():
            try:
                self._impl.save(key_state_pairs)
            except Exception as e:
                msg = "Failed to save state for %s: %s" % (keys, e)
                log.warning(msg)
                raise PersistenceStoreError(msg)

    def cleanup(self):
        self._save_from_buffer()
        self._impl.cleanup()

    @contextmanager
    def _timeit(self):
        """Log the time spent saving the state."""
        start_time = time.time()
        yield
        duration = time.time() - start_time
        log.info("State saved using %s in %0.3fs." % (self._impl, duration))

    @contextmanager
    def disabled(self):
        """Temporarily disable the state manager."""
        self.enabled, prev_enabled = False, self.enabled
        try:
            yield
        finally:
            self.enabled = prev_enabled


class NullStateManager(object):
    enabled = False

    @staticmethod
    def cleanup():
        pass

    @classmethod
    def disabled(cls):
        return cls()

    def __enter__(self):
        return

    def __exit__(self, *args):
        return


class StateChangeWatcher(observer.Observer):
    """Observer of stateful objects."""

    def __init__(self):
        self.state_manager = NullStateManager
        self.config = None

    def update_from_config(self, state_config):
        if self.config == state_config:
            return False

        self.shutdown()
        self.state_manager = PersistenceManagerFactory.from_config(
            state_config,
        )
        self.config = state_config
        return True

    def handler(self, observable, _event):
        """Handle a state change in an observable by saving its state."""
        if isinstance(observable, job.Job):
            self.save_job(observable)
        elif observable == MesosClusterRepository:
            self.save_frameworks(observable)

    def save_job(self, job):
        self._save_object(runstate.JOB_STATE, job)

    def save_frameworks(self, clusters):
        self._save_object(runstate.MESOS_STATE, clusters)

    def save_metadata(self):
        self._save_object(runstate.MCP_STATE, StateMetadata())

    def _save_object(self, state_type, obj):
        self.state_manager.save(state_type, obj.name, obj.state_data)

    def shutdown(self):
        self.state_manager.enabled = False
        self.state_manager.cleanup()

    def disabled(self):
        return self.state_manager.disabled()

    def restore(self, jobs):
        return self.state_manager.restore(jobs)



import datetime

import mock

from testifycompat import assert_equal
from testifycompat import setup
from testifycompat import TestCase
from tests import testingutils
from tests.assertions import assert_length
from tron import actioncommand
from tron.core import job
from tron.core.actionrun import ActionRun
from tron.core.job_scheduler import JobScheduler
from tron.core.job_scheduler import JobSchedulerFactory


class TestJobSchedulerGetRunsToSchedule(TestCase):
    @setup
    def setup_job(self):
        self.scheduler = mock.Mock()
        run_collection = mock.Mock(has_pending=False)
        node_pool = mock.Mock()
        self.job = job.Job(
            "jobname",
            self.scheduler,
            run_collection=run_collection,
            node_pool=node_pool,
        )
        self.job_scheduler = JobScheduler(self.job)
        self.job.runs.get_pending.return_value = False
        self.scheduler.queue_overlapping = True

    def test_get_runs_to_schedule_with_pending(self):
        self.scheduler.queue_overlapping = False
        self.job.runs.has_pending = True
        job_runs = self.job_scheduler.get_runs_to_schedule(None)
        assert_length(job_runs, 0)

    def test_get_runs_to_schedule_guess(self):
        job_runs = list(self.job_scheduler.get_runs_to_schedule(None))

        assert self.job.scheduler.next_run_time.call_args_list == [mock.call(None)]
        assert_length(job_runs, 1)
        # This should return a JobRun which has the job attached as an observer
        job_runs[0].attach.assert_any_call(True, self.job)

    def test_get_runs_to_schedule_given(self):
        now = datetime.datetime.now()
        job_runs = list(self.job_scheduler.get_runs_to_schedule(now))

        assert self.job.scheduler.next_run_time.call_count == 0
        assert_length(job_runs, 1)
        # This should return a JobRun which has the job attached as an observer
        job_runs[0].attach.assert_any_call(True, self.job)


class JobSchedulerManualStartTestCase(testingutils.MockTimeTestCase):

    now = datetime.datetime.now()

    @setup
    def setup_job(self):
        self.scheduler = mock.Mock()
        run_collection = mock.Mock()
        node_pool = mock.Mock()
        self.job = job.Job(
            "jobname",
            self.scheduler,
            run_collection=run_collection,
            node_pool=node_pool,
        )
        self.job_scheduler = JobScheduler(self.job)
        self.manual_run = mock.Mock()
        self.job.build_new_runs = mock.Mock(return_value=[self.manual_run])

    def test_manual_start(self):
        manual_runs = self.job_scheduler.manual_start()

        self.job.build_new_runs.assert_called_with(self.now, manual=True)
        assert_length(manual_runs, 1)
        self.manual_run.start.assert_called_once_with()

    def test_manual_start_default_with_timezone(self):
        self.job.time_zone = mock.Mock()
        with mock.patch(
            'tron.core.job_scheduler.timeutils.current_time',
            autospec=True,
        ) as mock_current:
            manual_runs = self.job_scheduler.manual_start()
            mock_current.assert_called_with(tz=self.job.time_zone)
            self.job.build_new_runs.assert_called_with(
                mock_current.return_value,
                manual=True,
            )
        assert_length(manual_runs, 1)
        self.manual_run.start.assert_called_once_with()

    def test_manual_start_with_run_time(self):
        run_time = datetime.datetime(2012, 3, 14, 15, 9, 26)
        manual_runs = self.job_scheduler.manual_start(run_time)

        self.job.build_new_runs.assert_called_with(run_time, manual=True)
        assert_length(manual_runs, 1)
        self.manual_run.start.assert_called_once_with()


class TestJobSchedulerSchedule(TestCase):
    @setup
    def setup_job(self):
        self.scheduler = mock.Mock(autospec=True)
        self.scheduler.next_run_time.return_value = 0
        mock_run = mock.Mock()
        mock_run.seconds_until_run_time.return_value = 0
        run_collection = mock.Mock(
            has_pending=False,
            autospec=True,
            return_value=[mock_run],
        )
        mock_build_new_run = mock.Mock()
        run_collection.build_new_run.return_value = mock_build_new_run
        mock_build_new_run.seconds_until_run_time.return_value = 0
        node_pool = mock.Mock()
        self.job = job.Job(
            name="jobname",
            scheduler=self.scheduler,
            run_collection=run_collection,
            node_pool=node_pool,
        )
        self.job_scheduler = JobScheduler(self.job)
        self.original_build_new_runs = self.job.build_new_runs
        self.job.build_new_runs = mock.Mock(return_value=[mock_run])

    @mock.patch('tron.core.job_scheduler.reactor', autospec=True)
    def test_enable(self, reactor):
        self.job.enabled = False
        self.job_scheduler.enable()
        assert self.job.enabled
        assert_length(reactor.callLater.mock_calls, 1)

    @mock.patch('tron.core.job_scheduler.reactor', autospec=True)
    def test_enable_noop(self, reactor):
        self.job.enabled = True
        self.job_scheduler.enable()
        assert self.job.enabled
        assert_length(reactor.callLater.mock_calls, 0)

    @mock.patch('tron.core.job_scheduler.reactor', autospec=True)
    def test_schedule(self, reactor):
        self.job.build_new_runs = self.original_build_new_runs
        self.job_scheduler.schedule()
        assert reactor.callLater.call_count == 1

        # Args passed to callLater
        call_args = reactor.callLater.mock_calls[0][1]
        assert_equal(call_args[1], self.job_scheduler.run_job)
        secs = call_args[0]
        run = call_args[2]

        run.seconds_until_run_time.assert_called_with()
        # Assert that we use the seconds we get from the run to schedule
        assert_equal(run.seconds_until_run_time.return_value, secs)

    @mock.patch('tron.core.job_scheduler.reactor', autospec=True)
    def test_schedule_disabled_job(self, reactor):
        self.job.enabled = False
        self.job_scheduler.schedule()
        assert reactor.callLater.call_count == 0

    @mock.patch('tron.core.job_scheduler.reactor', autospec=True)
    def test_handle_job_events_no_schedule_on_complete(self, reactor):
        self.job_scheduler.run_job = mock.Mock()
        self.job.scheduler.schedule_on_complete = False
        queued_job_run = mock.Mock()
        self.job.runs.get_first_queued = lambda: queued_job_run
        self.job_scheduler.handle_job_events(self.job, job.Job.NOTIFY_RUN_DONE)
        reactor.callLater.assert_any_call(
            0,
            self.job_scheduler.run_job,
            queued_job_run,
            run_queued=True,
        )

    def test_handle_job_events_schedule_on_complete(self):
        self.job_scheduler.schedule = mock.Mock()
        self.job.scheduler.schedule_on_complete = True
        self.job_scheduler.handle_job_events(self.job, job.Job.NOTIFY_RUN_DONE)
        self.job_scheduler.schedule.assert_called_with()

    def test_handler_unknown_event(self):
        self.job.runs.get_runs_by_state = mock.Mock()
        self.job_scheduler.handler(self.job, 'some_other_event')
        self.job.runs.get_runs_by_state.assert_not_called()

    def test_handler_no_queued(self):
        self.job_scheduler.run_job = mock.Mock()

        def get_queued(state):
            if state == ActionRun.QUEUED:
                return []

        self.job.runs.get_runs_by_state = get_queued
        self.job_scheduler.handler(self.job, job.Job.NOTIFY_RUN_DONE)
        self.job_scheduler.run_job.assert_not_called()

    @mock.patch('tron.core.job_scheduler.reactor', autospec=True)
    def test_run_queue_schedule(self, reactor):
        with mock.patch.object(
            self.job_scheduler,
            'schedule',
        ) as mock_schedule:
            self.job_scheduler.run_job = mock.Mock()
            self.job.scheduler.schedule_on_complete = False
            queued_job_run = mock.Mock()
            self.job.runs.get_first_queued = lambda: queued_job_run
            self.job_scheduler.run_queue_schedule()
            reactor.callLater.assert_called_once_with(
                0,
                self.job_scheduler.run_job,
                queued_job_run,
                run_queued=True,
            )
            mock_schedule.assert_called_once_with()


class TestJobSchedulerOther(TestCase):
    """ Test other JobScheduler functions """

    def _make_job_scheduler(self, job_name, enabled=True):
        scheduler = mock.Mock()
        run_collection = mock.Mock()
        node_pool = mock.Mock()
        new_job = job.Job(
            job_name,
            scheduler,
            run_collection=run_collection,
            node_pool=node_pool,
            enabled=enabled,
        )
        return new_job, JobScheduler(new_job)

    @setup
    def setup_job(self):
        self.job, self.job_scheduler = self._make_job_scheduler(
            'jobname', True
        )

    def test_disable(self):
        self.job.runs.cancel_pending = mock.Mock()

        self.job_scheduler.disable()

        assert not self.job.enabled
        assert self.job.runs.cancel_pending.call_count == 1

    def test_update_from_job_scheduler_disable(self):
        new_job, new_job_scheduler = self._make_job_scheduler('jobname', False)
        self.job.update_from_job = mock.Mock()
        self.job_scheduler.disable = mock.Mock()

        self.job_scheduler.update_from_job_scheduler(new_job_scheduler)

        assert self.job.update_from_job.call_args == mock.call(
            new_job_scheduler.get_job()
        )
        assert self.job_scheduler.disable.call_count == 1

    def test_update_from_job_scheduler_enable(self):
        new_job, new_job_scheduler = self._make_job_scheduler('jobname', True)
        self.job.update_from_job = mock.Mock()
        self.job.enabled = False
        self.job.config_enabled = False
        self.job_scheduler.enable = mock.Mock()

        self.job_scheduler.update_from_job_scheduler(new_job_scheduler)

        assert self.job.update_from_job.call_args == mock.call(
            new_job_scheduler.get_job()
        )
        assert self.job_scheduler.enable.call_count == 1

    def test_update_from_job_scheduler_no_config_change(self):
        new_job, new_job_scheduler = self._make_job_scheduler('jobname', True)
        self.job.enabled = False
        self.job.update_from_job = mock.Mock()
        self.job_scheduler.enable = mock.Mock()
        self.job_scheduler.disable = mock.Mock()

        self.job_scheduler.update_from_job_scheduler(new_job_scheduler)

        assert self.job.update_from_job.call_args == mock.call(
            new_job_scheduler.get_job()
        )
        assert self.job_scheduler.enable.call_count == 0
        assert self.job_scheduler.disable.call_count == 0
        assert self.job.config_enabled == new_job.config_enabled
        assert not self.job.enabled


class TestJobSchedulerFactory(TestCase):
    @setup
    def setup_factory(self):
        self.context = mock.Mock()
        self.output_stream_dir = mock.Mock()
        self.time_zone = mock.Mock()
        self.action_runner = mock.create_autospec(
            actioncommand.SubprocessActionRunnerFactory,
        )
        self.factory = JobSchedulerFactory(
            self.context,
            self.output_stream_dir,
            self.time_zone,
            self.action_runner,
            mock.Mock(),
        )

    def test_build(self):
        config = mock.Mock()
        with mock.patch(
            'tron.core.job_scheduler.Job', autospec=True
        ) as mock_job:
            job_scheduler = self.factory.build(config)
            _, kwargs = mock_job.from_config.call_args
            assert_equal(kwargs['job_config'], config)
            assert_equal(
                job_scheduler.get_job(),
                mock_job.from_config.return_value,
            )
            assert_equal(kwargs['parent_context'], self.context)
            assert_equal(kwargs['output_path'].base, self.output_stream_dir)
            assert_equal(kwargs['action_runner'], self.action_runner)



#!/usr/bin/env python3.6
import argparse
import logging
import signal
import sys
from queue import Queue

import psutil
from twisted.internet import inotify
from twisted.internet import reactor
from twisted.python import filepath

from tron import yaml

log = logging.getLogger('tron.recover_batch')


class StatusFileWatcher(object):
    """
    Watches the status file produced by action runners
    """

    def __init__(self, to_watch, callback):
        notifier = inotify.INotify()
        notifier.startReading()
        notifier.watch(filepath.FilePath(to_watch), callbacks=[callback])


def parse_args():
    parser = argparse.ArgumentParser(
        description='Check if a action runner has exited; wait otherwise',
    )
    parser.add_argument('filepath', )
    return parser.parse_args()


def read_last_yaml_entries(filename):
    with open(filename) as f:
        lines = list(yaml.load_all(f))
        if not lines:
            entries = {}
        else:
            entries = lines[-1]
    return entries


def notify(notify_queue, ignored, filepath, mask):
    exit_code, error_message = get_exit_code(filepath.path)
    if exit_code is not None:
        reactor.stop()
        notify_queue.put((exit_code, error_message))


def get_exit_code(filepath):
    entries = read_last_yaml_entries(filepath)
    pid = entries.get('runner_pid')
    return_code = entries.get('return_code')
    exit_code, error_message = None, None

    if return_code is not None:
        if return_code < 0:
            # from the subprocess docs on the return code of a process:
            # "A negative value -N indicates that the child was terminated by signal N (POSIX only)."
            # We should always exit with a positive code, so we take the absolute value of the return code
            exit_code = abs(return_code)
            error_message = f'Action run killed by signal {signal.Signals(exit_code).name}'
        else:
            exit_code = return_code
    elif pid is None:
        log.warning(f"Status file {filepath} didn't have a PID. Will watch the file for updates.")
    elif not psutil.pid_exists(pid):
        exit_code = 1
        error_message = f'Action runner pid {pid} no longer running. Assuming an exit of 1.'

    return exit_code, error_message


def run(fpath):
    # Check if the process has already completed.
    # If it has, we don't expect any more updates.
    return_code, error_message = get_exit_code(fpath)
    if return_code is not None:
        if error_message is not None:
            log.warning(error_message)
        sys.exit(return_code)

    # If not, wait for updates to the file.
    notify_queue = Queue()
    StatusFileWatcher(
        fpath,
        lambda *args, **kwargs: notify(notify_queue, *args, **kwargs)
    )
    reactor.run()
    exit_code, error_message = notify_queue.get()
    if error_message is not None:
        log.warning(error_message)
    sys.exit(exit_code)


if __name__ == "__main__":
    args = parse_args()
    run(args.filepath)



import mock
import pytest

from tron import mcp
from tron.api import controller
from tron.api.controller import ConfigController
from tron.api.controller import EventsController
from tron.api.controller import JobCollectionController
from tron.api.controller import UnknownCommandError
from tron.config import ConfigError
from tron.config import manager
from tron.core import actionrun
from tron.core import jobrun
from tron.core.job_collection import JobCollection
from tron.core.job_scheduler import JobScheduler


class TestJobCollectionController:
    @pytest.fixture(autouse=True)
    def setup_controller(self):
        self.collection = mock.create_autospec(
            JobCollection,
            enable=mock.Mock(),
            disable=mock.Mock(),
        )
        self.controller = JobCollectionController(self.collection)

    def test_handle_command_unknown(self):
        with pytest.raises(UnknownCommandError):
            self.controller.handle_command('enableall')
            self.controller.handle_command('disableall')

    def test_handle_command_move_non_existing_job(self):
        self.collection.get_names.return_value = []
        result = self.controller.handle_command(
            'move', old_name='old.test', new_name='new.test'
        )
        assert "doesn't exist" in result

    def test_handle_command_move_to_existing_job(self):
        self.collection.get_names.return_value = ['old.test', 'new.test']
        result = self.controller.handle_command(
            'move', old_name='old.test', new_name='new.test'
        )
        assert "exists already" in result

    def test_handle_command_move(self):
        self.collection.get_names.return_value = ['old.test']
        result = self.controller.handle_command(
            'move', old_name='old.test', new_name='new.test'
        )
        assert "Error" not in result


class TestActionRunController:
    @pytest.fixture(autouse=True)
    def setup_controller(self):
        self.action_run = mock.create_autospec(
            actionrun.ActionRun,
            cancel=mock.Mock(),
        )
        self.job_run = mock.create_autospec(jobrun.JobRun)
        self.job_run.is_scheduled = False
        self.controller = controller.ActionRunController(
            self.action_run,
            self.job_run,
        )

    def test_handle_command_start_failed(self):
        self.job_run.is_scheduled = True
        result = self.controller.handle_command('start')
        assert not self.action_run.start.mock_calls
        assert "cannot be started" in result

    def test_handle_command_recover_failed(self):
        self.action_run.is_unknown = False
        result = self.controller.handle_command('recover')
        assert not self.action_run.recover.mock_calls
        assert "cannot be recovered" in result

    def test_handle_command_mapped_command(self):
        result = self.controller.handle_command('cancel')
        self.action_run.cancel.assert_called_with()
        assert "now in state" in result

    def test_handle_command_mapped_command_failed(self):
        self.action_run.cancel.return_value = False
        result = self.controller.handle_command('cancel')
        self.action_run.cancel.assert_called_with()
        assert "Failed to cancel" in result

    def test_handle_termination_not_implemented(self):
        self.action_run.stop.side_effect = NotImplementedError
        result = self.controller.handle_termination('stop')
        assert "Failed to stop" in result

    def test_handle_termination_success_without_extra_msg(self):
        self.action_run.kill.return_value = None
        result = self.controller.handle_termination('kill')
        assert "Attempting to kill" in result

    def test_handle_termination_success_with_extra_msg(self):
        self.action_run.kill.return_value = "Warning Message"
        result = self.controller.handle_termination('kill')
        assert "Attempting to kill" in result
        assert "Warning Message" in result


class TestJobRunController:
    @pytest.fixture(autouse=True)
    def setup_controller(self):
        self.job_run = mock.create_autospec(
            jobrun.JobRun,
            run_time=mock.Mock(),
            cancel=mock.Mock(),
        )
        self.job_scheduler = mock.create_autospec(JobScheduler)
        self.controller = controller.JobRunController(
            self.job_run,
            self.job_scheduler,
        )

    def test_handle_command_restart(self):
        self.controller.handle_command('restart')
        self.job_scheduler.manual_start.assert_called_with(
            self.job_run.run_time,
        )

    def test_handle_mapped_command(self):
        result = self.controller.handle_command('start')
        self.job_run.start.assert_called_with()
        assert 'now in state' in result

    def test_handle_mapped_command_failure(self):
        self.job_run.cancel.return_value = False
        result = self.controller.handle_command('cancel')
        self.job_run.cancel.assert_called_with()
        assert 'Failed to cancel' in result


class TestJobController:
    @pytest.fixture(autouse=True)
    def setup_controller(self):
        self.job_scheduler = mock.create_autospec(JobScheduler)
        self.controller = controller.JobController(self.job_scheduler)

    def test_handle_command_enable(self):
        self.controller.handle_command('enable')
        self.job_scheduler.enable.assert_called_with()

    def test_handle_command_disable(self):
        self.controller.handle_command('disable')
        self.job_scheduler.disable.assert_called_with()

    def test_handle_command_start(self):
        run_time = mock.Mock()
        self.controller.handle_command('start', run_time)
        self.job_scheduler.manual_start.assert_called_with(run_time=run_time)


class TestConfigController:
    @pytest.fixture(autouse=True)
    def setup_controller(self):
        self.mcp = mock.create_autospec(mcp.MasterControlProgram)
        self.manager = mock.create_autospec(manager.ConfigManager)
        self.mcp.get_config_manager.return_value = self.manager
        self.controller = ConfigController(self.mcp)

    def test_get_config_content_new(self):
        self.manager.__contains__.return_value = False
        content = self.controller._get_config_content('name')
        assert content == self.controller.DEFAULT_NAMED_CONFIG
        assert not self.manager.read_raw_config.call_count

    def test_get_config_content_old(self):
        self.manager.__contains__.return_value = True
        name = 'the_name'
        content = self.controller._get_config_content(name)
        assert content == self.manager.read_raw_config.return_value
        self.manager.read_raw_config.assert_called_with(name)

    def test_read_config(self):
        self.manager.__contains__.return_value = True
        name = 'MASTER'
        resp = self.controller.read_config(name)
        self.manager.read_raw_config.assert_called_with(name)
        self.manager.get_hash.assert_called_with(name)
        assert resp['config'] == self.manager.read_raw_config.return_value
        assert resp['hash'] == self.manager.get_hash.return_value

    def test_update_config(self):
        name, content, config_hash = None, mock.Mock(), mock.Mock()
        self.manager.get_hash.return_value = config_hash
        assert not self.controller.update_config(name, content, config_hash)
        self.manager.get_hash.assert_called_with(name)
        self.manager.write_config.assert_called_with(name, content)
        self.mcp.reconfigure.assert_called_with()

    def test_update_config_failure(self):
        name, content, old_content, config_hash = None, mock.Mock(), mock.Mock(), mock.Mock()
        self.manager.get_hash.return_value = config_hash
        self.manager.write_config.side_effect = [ConfigError("It broke"), None]
        self.controller.read_config = mock.Mock(return_value={'config': old_content})
        error = self.controller.update_config(
            name,
            content,
            config_hash,
        )
        assert error == "It broke"
        self.manager.write_config.call_args_list = [(name, content), (name, old_content)]
        assert self.mcp.reconfigure.call_count == 1

    def test_update_config_hash_mismatch(self):
        name, content, config_hash = None, mock.Mock(), mock.Mock()
        error = self.controller.update_config(name, content, config_hash)
        assert error == "Configuration has changed. Please try again."

    def test_delete_config(self):
        name, content, config_hash = None, "", mock.Mock()
        self.manager.get_hash.return_value = config_hash
        assert not self.controller.delete_config(name, content, config_hash)
        self.manager.delete_config.assert_called_with(name)
        self.mcp.reconfigure.assert_called_with()
        self.manager.get_hash.assert_called_with(name)

    def test_delete_config_failure(self):
        name, content, config_hash = None, "", mock.Mock()
        self.manager.get_hash.return_value = config_hash
        self.manager.delete_config.side_effect = Exception("some error")
        error = self.controller.delete_config(name, content, config_hash)
        assert error
        self.manager.delete_config.assert_called_with(name)
        assert not self.mcp.reconfigure.call_count

    def test_delete_config_hash_mismatch(self):
        name, content, config_hash = None, "", mock.Mock()
        error = self.controller.delete_config(name, content, config_hash)
        assert error == "Configuration has changed. Please try again."

    def test_delete_config_content_not_empty(self):
        name, content, config_hash = None, "content", mock.Mock()
        error = self.controller.delete_config(name, content, config_hash)
        assert error

    def test_get_namespaces(self):
        result = self.controller.get_namespaces()
        self.manager.get_namespaces.assert_called_with()
        assert result == self.manager.get_namespaces.return_value


class TestEventsController:
    @pytest.fixture(autouse=True)
    def setup(self):
        with mock.patch('tron.api.controller.EventBus', autospec=True) as eb:
            eb.instance = mock.Mock()
            self.eventbus = eb
            self.controller = EventsController()
            yield

    def test_info(self):
        self.eventbus.instance = None
        assert self.controller.info() == dict(error='EventBus disabled')

        self.eventbus.instance = mock.Mock()
        assert self.controller.info() == dict(response=self.eventbus.instance.event_log)

    def test_publish(self):
        event = mock.Mock()
        self.eventbus.instance = None
        self.eventbus.has_event.return_value = True
        self.eventbus.publish.return_value = False

        assert self.controller.info() == dict(error='EventBus disabled')
        assert len(self.eventbus.publish.mock_calls) == 0

        self.eventbus.instance = mock.Mock()
        assert self.controller.publish(event) == dict(response=f'event {event} already published')
        assert len(self.eventbus.publish.mock_calls) == 0

        self.eventbus.has_event.return_value = False
        assert self.controller.publish(event) == dict(error=f'could not publish {event}')
        assert len(self.eventbus.publish.mock_calls) == 1

        self.eventbus.publish.return_value = True
        assert self.controller.publish(event) == dict(response=f'OK')
        assert len(self.eventbus.publish.mock_calls) == 2

    def test_discard(self):
        event = mock.Mock()
        self.eventbus.instance = None
        self.eventbus.discard.return_value = False

        assert self.controller.info() == dict(error='EventBus disabled')
        assert len(self.eventbus.discard.mock_calls) == 0

        self.eventbus.instance = mock.Mock()
        assert self.controller.discard(event) == dict(error=f'could not discard {event}')
        assert len(self.eventbus.discard.mock_calls) == 1

        self.eventbus.discard.return_value = True
        assert self.controller.discard(event) == dict(response=f'OK')
        assert len(self.eventbus.discard.mock_calls) == 2



#!/usr/bin/env python3.6
"""
Write pid and stdout/stderr to a standard location before execing a command.
"""
import argparse
import contextlib
import logging
import os
import subprocess
import sys
import threading
import time

from tron import yaml

STATUS_FILE = 'status'


class StatusFile(object):
    """Manage a status file."""

    def __init__(self, filename):
        self.filename = filename

    def get_content(self, run_id, command, proc):
        return {
            'run_id': run_id,
            'command': command,
            'pid': proc.pid,
            'return_code': proc.returncode,
            'runner_pid': os.getpid(),
            'timestamp': time.time(),
        }

    @contextlib.contextmanager
    def wrap(self, command, run_id, proc):
        with open(self.filename, 'w') as fh:
            yaml.safe_dump(
                self.get_content(
                    run_id=run_id,
                    command=command,
                    proc=proc,
                ),
                fh,
                explicit_start=True,
                width=1000000,
            )
        try:
            yield
        finally:
            with open(self.filename, 'a') as fh:
                yaml.safe_dump(
                    self.get_content(
                        run_id=run_id,
                        command=command,
                        proc=proc,
                    ),
                    fh,
                    explicit_start=True,
                    width=1000000,
                )


def validate_output_dir(path):
    if os.path.isdir(path):
        if not os.access(path, os.W_OK):
            raise OSError("Output dir %s not writable" % path)
        return
    else:
        try:
            os.makedirs(path)
        except OSError:
            raise OSError("Could not create output dir %s" % path)


def build_environment(run_id, original_env=None):
    if original_env is None:
        original_env = dict(os.environ)
    try:
        namespace, job, run_num, action = run_id.split('.', maxsplit=3)
    except ValueError:
        # if we can't parse the run_id, we don't want to abort, so just
        # set these semi-arbitrarily
        namespace, job, run_num, action = ['UNKNOWN'] * 4

    new_env = dict(original_env)
    new_env['TRON_JOB_NAMESPACE'] = namespace
    new_env['TRON_JOB_NAME'] = job
    new_env['TRON_RUN_NUM'] = run_num
    new_env['TRON_ACTION'] = action

    logging.debug(new_env)
    return new_env


def run_proc(output_path, command, run_id, proc):
    logging.warning(f'{run_id} running as pid {proc.pid}')
    status_file = StatusFile(os.path.join(output_path, STATUS_FILE))
    with status_file.wrap(
        command=command,
        run_id=run_id,
        proc=proc,
    ):
        returncode = proc.wait()
        logging.warning(f'pid {proc.pid} exited with returncode {returncode}')
        return returncode


def parse_args():
    parser = argparse.ArgumentParser(description='Action Runner for Tron')
    parser.add_argument(
        'output_dir',
        help='The directory to store the state of the action run',
    )
    parser.add_argument(
        'command',
        help='the command to run',
    )
    parser.add_argument(
        'run_id',
        help='run_id of the action',
    )
    return parser.parse_args()


def run_command(command, run_id):
    return subprocess.Popen(
        command,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        env=build_environment(run_id=run_id),
    )


def stream(source, dst):
    is_connected = True
    logging.warning(f'streaming {source.name} to {dst.name}')
    for line in iter(source.readline, b''):
        if is_connected:
            try:
                dst.write(line.decode('utf-8'))
                dst.flush()
                logging.warning(f'{dst.name}: {line}')
            except Exception as e:
                logging.warning(f'failed writing to {dst}: {e}')
                logging.warning(f'{dst.name}: {line}')
                is_connected = False
        else:
            logging.warning(f'{dst.name}: {line}')
            is_connected = False


def configure_logging(run_id, output_dir):
    output_file = os.path.join(output_dir, f'{run_id}-{os.getpid()}.log')
    logging.basicConfig(
        filename=output_file,
        format='%(asctime)s %(levelname)s %(message)s',
        datefmt='%Y-%m-%dT%H:%M:%S%z'
    )


def main():
    args = parse_args()
    validate_output_dir(args.output_dir)
    configure_logging(run_id=args.run_id, output_dir=args.output_dir)
    proc = run_command(command=args.command, run_id=args.run_id)
    threads = [
        threading.Thread(target=stream, args=p, daemon=True)
        for p in [(proc.stdout, sys.stdout), (proc.stderr, sys.stderr)]
    ]
    for t in threads:
        t.start()
    returncode = run_proc(
        output_path=args.output_dir,
        run_id=args.run_id,
        command=args.command,
        proc=proc,
    )

    for t in threads:
        t.join()

    return returncode


if __name__ == "__main__":
    sys.exit(main())



from __future__ import absolute_import
from __future__ import unicode_literals

import logging
from collections import namedtuple

from tron.utils.timeutils import delta_total_seconds

log = logging.getLogger(__name__)
Trigger = namedtuple('Trigger', ['name', 'command'])


class ActionGraph(object):
    """A directed graph of actions and their requirements for a specific job."""

    def __init__(self, action_map, required_actions, required_triggers):
        self.action_map = action_map
        self.required_actions = required_actions
        self.required_triggers = required_triggers
        self.all_triggers = set(self.required_triggers)
        for action_triggers in self.required_triggers.values():
            self.all_triggers |= action_triggers
        self.all_triggers -= set(self.action_map)

    def get_dependencies(self, action_name, include_triggers=False):
        """Given an Action's name return the Actions (and optionally, Triggers)
        required to run before that Action.
        """
        if action_name not in set(self.action_map) | self.all_triggers:
            return []

        dependencies = [
            self.action_map[action]
            for action in self.required_actions[action_name]
        ]
        if include_triggers:
            dependencies += [
                self[trigger_name]
                for trigger_name in self.required_triggers[action_name]
            ]
        return dependencies

    def names(self, include_triggers=False):
        names = set(self.action_map)
        if include_triggers:
            names |= self.all_triggers
        return names

    @property
    def expected_runtime(self):
        return {
            name: delta_total_seconds(self.action_map[name].expected_runtime)
            for name in self.action_map.keys()
        }

    def __getitem__(self, name):
        if name in self.action_map:
            return self.action_map[name]
        elif name in self.all_triggers:
            # we don't have the Trigger config to know what the real command is,
            # so we just fill in the command with 'TRIGGER'
            return Trigger(name, 'TRIGGER')
        else:
            raise KeyError(f'{name} is not a valid action')

    def __eq__(self, other):
        return (
            self.action_map == other.action_map and
            self.required_actions == other.required_actions and
            self.required_triggers == other.required_triggers
        )

    def __ne__(self, other):
        return not self == other



"""
Tron schedulers

 A scheduler has a simple interface.

 class Scheduler(object):

    schedule_on_complete = <bool>

    def next_run_time(self, last_run_time):
        <returns datetime>


 next_run_time() should return a datetime which is the time the next job run
 will be run.

 schedule_on_complete is a bool that identifies if this scheduler should have
 jobs scheduled with the start_time of the previous run (False), or the
 end time of the previous run (False).
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import datetime
import logging
import random

from tron.config import schedule_parse
from tron.utils import timeutils
from tron.utils import trontimespec

log = logging.getLogger(__name__)


def scheduler_from_config(config, time_zone):
    """A factory for creating a scheduler from a configuration object."""
    if isinstance(config, schedule_parse.ConfigGrocScheduler):
        return GeneralScheduler(
            time_zone=time_zone,
            timestr=config.timestr or '00:00',
            ordinals=config.ordinals,
            monthdays=config.monthdays,
            months=config.months,
            weekdays=config.weekdays,
            name='groc',
            original=config.original,
            jitter=config.jitter,
        )

    if isinstance(config, schedule_parse.ConfigCronScheduler):
        return GeneralScheduler(
            time_zone=time_zone,
            minutes=config.minutes,
            hours=config.hours,
            monthdays=config.monthdays,
            months=config.months,
            weekdays=config.weekdays,
            ordinals=config.ordinals,
            seconds=[0],
            name='cron',
            original=config.original,
            jitter=config.jitter,
        )

    if isinstance(config, schedule_parse.ConfigDailyScheduler):
        return GeneralScheduler(
            hours=[config.hour],
            time_zone=time_zone,
            minutes=[config.minute],
            seconds=[config.second],
            weekdays=config.days,
            name='daily',
            original=config.original,
            jitter=config.jitter,
        )


def get_jitter(time_delta):
    if not time_delta:
        return datetime.timedelta()
    seconds = timeutils.delta_total_seconds(time_delta)
    return datetime.timedelta(seconds=random.randint(-seconds, seconds))


def get_jitter_str(time_delta):
    if not time_delta:
        return ''
    return ' (+/- %s)' % time_delta


class GeneralScheduler(object):
    """Scheduler which uses a TimeSpecification.
    """
    schedule_on_complete = False

    def __init__(
        self,
        ordinals=None,
        weekdays=None,
        months=None,
        monthdays=None,
        timestr=None,
        minutes=None,
        hours=None,
        seconds=None,
        time_zone=None,
        name=None,
        original=None,
        jitter=None,
    ):
        """Parameters:
          timestr     - the time of day to run, as 'HH:MM'
          ordinals    - first, second, third &c, as a set of integers in 1..5 to
                        be used with "1st <weekday>", etc.
          monthdays   - set of integers to be used with "<month> 3rd", etc.
          months      - the months that this should run, as a set of integers in
                        1..12
          weekdays    - the days of the week that this should run, as a set of
                        integers, 0=Sunday, 6=Saturday
          timezone    - the optional timezone as a string for this specification.
                        Defaults to UTC - valid entries are things like
                        Australia/Victoria or PST8PDT.
        """
        self.time_zone = time_zone
        self.jitter = jitter
        self.name = name or 'daily'
        self.original = original or ''
        self.time_spec = trontimespec.TimeSpecification(
            ordinals=ordinals,
            weekdays=weekdays,
            months=months,
            monthdays=monthdays,
            timestr=timestr,
            hours=hours,
            minutes=minutes,
            seconds=seconds,
            timezone=time_zone.zone if time_zone else None,
        )

    def next_run_time(self, start_time):
        """Find the next time to run."""
        if not start_time:
            start_time = timeutils.current_time(tz=self.time_zone)
        elif self.time_zone:
            if start_time.tzinfo is None or start_time.tzinfo.utcoffset(
                start_time,
            ) is None:
                # tz-naive start times need to be localized first to the requested
                # time zone.
                start_time = trontimespec.naive_as_timezone(start_time, self.time_zone)

        return self.time_spec.get_match(start_time) + get_jitter(self.jitter)

    def __str__(self):
        return '%s %s%s' % (
            self.name,
            self.original,
            get_jitter_str(self.jitter),
        )

    def __eq__(self, other):
        return hasattr(
            other,
            'time_spec',
        ) and self.time_spec == other.time_spec

    def __ne__(self, other):
        return not self == other

    def get_jitter(self):
        return self.jitter

    def get_name(self):
        return self.name

    def get_value(self):
        return self.original



from __future__ import absolute_import
from __future__ import unicode_literals

import datetime
from unittest.mock import MagicMock

from testifycompat import assert_equal
from testifycompat import run
from testifycompat import setup
from testifycompat import TestCase
from tron.api.requestargs import get_bool
from tron.api.requestargs import get_datetime
from tron.api.requestargs import get_integer
from tron.api.requestargs import get_string


class TestRequestArgs(TestCase):
    @setup
    def setup_args(self):
        self.args = {
            b'number': [b'123'],
            b'string': [b'astring'],
            b'boolean': [b'1'],
            b'datetime': [b'2012-03-14 15:09:26'],
        }
        self.datetime = datetime.datetime(2012, 3, 14, 15, 9, 26)
        self.request = MagicMock(args=self.args)

    def _add_arg(self, name, value):
        name = name.encode()
        value = value.encode()
        if name not in self.args:
            self.args[name] = []
        self.args[name].append(value)

    def test_get_integer_valid_int(self):
        self._add_arg('number', '5')
        assert_equal(get_integer(self.request, 'number'), 123)

    def test_get_integer_invalid_int(self):
        self._add_arg('nan', 'beez')
        assert not get_integer(self.request, 'nan')

    def test_get_integer_missing(self):
        assert not get_integer(self.request, 'missing')

    def test_get_string(self):
        self._add_arg('string', 'bogus')
        assert_equal(get_string(self.request, 'string'), 'astring')

    def test_get_string_missing(self):
        assert not get_string(self.request, 'missing')

    def test_get_bool(self):
        assert get_bool(self.request, 'boolean')

    def test_get_bool_false(self):
        self._add_arg('false', '0')
        assert not get_bool(self.request, 'false')

    def test_get_bool_missing(self):
        assert not get_bool(self.request, 'missing')

    def test_get_datetime_valid(self):
        assert_equal(get_datetime(self.request, 'datetime'), self.datetime)

    def test_get_datetime_invalid(self):
        self._add_arg('nope', '2012-333-4')
        assert not get_datetime(self.request, 'nope')

    def test_get_datetime_missing(self):
        assert not get_datetime(self.request, 'missing')


if __name__ == "__main__":
    run()



from unittest import mock

from testifycompat import assert_equal
from testifycompat import assert_raises
from testifycompat import run
from testifycompat import setup
from testifycompat import TestCase
from tron.core import actiongraph


class TestActionGraph(TestCase):
    @setup
    def setup_graph(self):
        self.action_names = [
            'base_one',
            'base_two',
            'dep_one',
            'dep_one_one',
            'dep_multi',
        ]
        self.action_map = {}
        for name in self.action_names:
            self.action_map[name] = mock.MagicMock()
            self.action_map[name].name = name

        self.required_actions = {
            'base_one': set(),
            'base_two': set(),
            'dep_multi': {'dep_one_one', 'base_two'},
            'dep_one_one': {'dep_one'},
            'dep_one': {'base_one'},
        }
        self.required_triggers = {
            'base_one': {'MASTER.otherjob.first'},
            'base_two': set(),
            'dep_multi': set(),
            'dep_one_one': set(),
            'dep_one': set(),
        }

        self.action_graph = actiongraph.ActionGraph(self.action_map, self.required_actions, self.required_triggers)

    def test_get_dependencies(self):
        assert self.action_graph.get_dependencies('not_in_job') == []
        assert self.action_graph.get_dependencies('base_one') == []
        assert self.action_graph.get_dependencies('base_one', include_triggers=True)[0].name == 'MASTER.otherjob.first'
        assert sorted([d.name for d in self.action_graph.get_dependencies('dep_multi')]) == sorted([
            'dep_one_one', 'base_two',
        ])

    def test_names(self):
        assert sorted(self.action_graph.names()) == sorted(self.action_names)
        assert sorted(self.action_graph.names(include_triggers=True)) == sorted(
            self.action_names + ['MASTER.otherjob.first']
        )

    def test__getitem__(self):
        assert_equal(
            self.action_graph['base_one'],
            self.action_map['base_one'],
        )

    def test__getitem__miss(self):
        assert_raises(KeyError, lambda: self.action_graph['unknown'])

    def test__eq__(self):
        other_graph = mock.MagicMock(
            action_map=self.action_map,
            required_actions=self.required_actions,
            required_triggers=self.required_triggers,
        )
        assert_equal(self.action_graph, other_graph)

        other_graph.required_actions = None
        assert not self.action_graph == other_graph

    def test__ne__(self):
        other_graph = mock.MagicMock()
        assert self.action_graph != other_graph


if __name__ == "__main__":
    run()



from __future__ import absolute_import
from __future__ import unicode_literals

from testifycompat import assert_equal
from testifycompat import setup
from testifycompat import TestCase
from tron.utils import state


class TestStateMachineSimple(TestCase):
    @setup
    def build_machine(self):
        self.state_green = 'green'
        self.state_red = 'red'

        self.machine = state.Machine(self.state_red, red=dict(true='green'))

    def test_transition_many(self):
        # Stay the same
        assert not self.machine.transition('missing')
        assert_equal(self.machine.state, self.state_red)

        # Traffic has arrived
        self.machine.transition('true')
        assert_equal(self.machine.state, self.state_green)

        # Still traffic
        self.machine.transition('true')
        assert_equal(self.machine.state, self.state_green)

    def test_check(self):
        assert not self.machine.check(False)
        assert_equal(self.machine.check('true'), self.state_green)
        assert_equal(self.machine.state, self.state_red)


class TestStateMachineMultiOption(TestCase):
    @setup
    def build_machine(self):
        # Generalized rules of a conversation
        # If they are talking, we should listen
        # If they are listening, we should talk
        # If they are ignoring us we should get angry
        self.machine = state.Machine(
            'listening',
            listening=dict(listening='talking'),
            talking=dict(ignoring='angry', talking='listening'),
        )

    def test_transition_many(self):
        # Talking, we should listen
        self.machine.transition("talking")
        assert_equal(self.machine.state, 'listening')

        # Now be polite
        self.machine.transition("listening")
        assert_equal(self.machine.state, 'talking')

        self.machine.transition("listening")
        assert_equal(self.machine.state, 'talking')

        # But they are tired of us...
        self.machine.transition("ignoring")
        assert_equal(self.machine.state, 'angry')

    def test_transition_set(self):
        expected = {'listening', 'talking', 'ignoring'}
        assert_equal(set(self.machine.transition_names), expected)
